/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_22.c - ? menus
*/
#include <stdio.h>
#include <smgdef.h>
#include <smgmsg.h>
#include <ssdef.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

#define MM_PASTE_Y	2
#define MM_PASTE_X	2
#define MM_ROWS		22
#define MM_COLS		78
#define MM_FIRST_LINE	1
#define MM_INDENT	1


static int menu_inited;
static int display_mm;

/*
**  we have 78 columns by 22 rows to do our stuff
**                                    ||
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

char *mm_choices[] = {
"",
" @      change file protection            ~   change file ownership",
" C      create a new directory            D   delete selected directories",
" !      show directory statistics         H   help",
#ifdef MATS_HACK
#ifdef Q_EXITS_TO_ORIGINAL_DIR
" E,[CR] exit to current default dir       Q   exit to original default dir",
"^E      edit a new file                  ^G   create list of curr dir struct",
" F      file manager (non-directories)    O   show broadcast messages",
#else
" E,Q,[CR] exit to current default dir     O   show broadcast messages",
"^E      edit a new file                  ^G   create a listing of the current",
" F      file manager (non-directories)        directory structure",
#endif
#else
#ifdef Q_EXITS_TO_ORIGINAL_DIR
" E      exit to current default dir       Q   exit to original default dir",
"^E      edit a new file                  ^G   create list of curr dir struct",
" F,[CR] file manager (non-directories)    O   show broadcast messages",
#else
" E,Q    exit to current default dir       O   show broadcast messages",
"^E      edit a new file                  ^G   create a listing of the current",
" F,[CR] file manager (non-directories)        directory structure",
#endif
#endif
"^F      file manager (directories only)   K   remove current directory",
" I      introduce another dir structure       structure from the display",
" L      locate a file                     M   move a directory",
" R      rename selected directory         S   rebuild directory structure",
" U      purge selected directories        W   show current default directory",
" V      show current SWING version       ^D   user definable options",
"^W,^R   repaint the screen                <   import files from another struc",
" &      show disk quota and summary       *   snapshot the screen",
"=/#     include/exclude files shown,     /[{( set mark  \\]}) return to mark",
"        e.g. *.c, *.obj or test*.*        %   defragment selected directories",
"",
"    N   next page               P   previous page             $   spawn DCL",
"    T   top of display          B   bottom of display         ?   this menu",
"",
"                    Select option or <RETURN> to continue",
0};

char *st_choices[] = {
"                A  -  blocks allocated to files in directory",
"                C  -  count of files in directory",
"                D  -  the directory name",
"                G  -  grand statistical summary",
"                F  -  file manager",
"                =  -  file manager with partial file specification",
"                I  -  fragmentation index",
"                L  -  list the directory structure together",
"                      with the statistical information",
"                M  -  the maximum file size",
"                Q  -  quit (see comment below)",
"                S  -  re-construct the statistical information",
"                U  -  blocks used by files in directory",
"",
"        You can optionally quit and leave the stats on display,",
"        thus allowing cursor movement. To re-set the display back",
"        to the display of directory names, type `!' again.",
"",
"        You can use the arrow keys, T(op), B(ottom), Next Screen",
"        or Prev Screen. This is the standard SWING cursor control.",
"",
"                    Select option or <RETURN> to continue",
0};

/*
**  we have 78 columns by 22 rows to do our stuff
**                                    ||
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

char *fm_choices[] = {
"",
" @      change file protection           ~    change file ownership",
" +/-    select/deselect files based on   D    delete selected files",
"        a partial file specification     E    edit current file",
"^D      user definable options          ^E    edit a new file",
" A/Z    select/deselect all files        C    copy selected files",
" F      reconstruct file list            H    help   G - op on sel'd files",
" L      list (print) selected files      M    move selected files to another",
" O      show broadcast messages               directory structure",
#ifdef CR_RETURNS_FROM_FILE_TO_DIR_MGR
"[CR],Q  quit the file manager            Y    type current/selected file(s)",  
#else
" Q      quit the file manager       [CR],Y    type current/selected file(s)",  
#endif
" S      submit selected files to batch   R    rename selected files", 
" *      snapshot the screen              U    purge this directory",  
"^W,^R   repaint the screen               X    execute current file",  
" </>    import/export files from       <- ->  change directory info format",
"        another file structure    [SPACE_BAR] select/deselect current file",
" =/#    include/exclude files shown,     %    defragment selected files",
"        e.g. *.c, *.obj or test*.*   [SELECT] next directory on the stack",
" &      show disk quota and summary   [FIND]  previous directory on the stack",
"",
" N    next page              P    previous page             $   spawn DCL",
" T    top file               B    last file                 ?   this menu",  
"",
"                    Select option or <RETURN> to continue",
0};

/*
**  we have 78 columns by 22 rows to do our stuff
**                                    ||
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

char *dm_choices[] = {
"",
"",
"                  ^D    user definable options",
"                   F    reconstruct file list",
"                   H    help",
"                   O    show broadcast messages",
"                   Q    quit the file manager",
"                  ^W,^R repaint the screen",
"                   $    DCL",
"                   &    show disk quota and summary",
"                   @    change file protection",
"                   *    snapshot the screen",
"",
"               <- ->    change directory info format",
"",
"",
"                 N    next page     P    previous page",
"                 T    top file      B    last file",  
"",
"",
"",
"                 Select option or <RETURN> to continue",
0};

/*
**  we have 78 columns by 22 rows to do our stuff
**                                    ||
123456789012345678901234567890123456789012345678901234567890123456789012345678*/

char *df_choices[] = {
"",
"              C      change the current definition",
"              R      restore original value of the current entry",
"              Q      quit and lose any changes",
"              E      exit and save changes",
"        [SPACE_BAR]  same as C",
"              *      snapshot the screen",
"",
"",
"        T      top of display            B      bottom of display",
"        N      next screen               P      previous screen",
"",
"                    up-arrow     up a line",
"                  down-arrow     down a line",
"",
"       N.B. terminate the line with ! if you don't want the current",
"       file name (where appropriate) to be appended to the command",
"",
"       N.B. you can then add an ^ if you don't want to be prompted",
"       to confirm the execution of the command you've defined for",
"       this key",
"",
"",
"                 Select option or <RETURN> to continue",
0};

init_menu()
{
    int status;

    menu_inited = TRUE;

    if ((status = smg$create_virtual_display(&MM_ROWS,
                                             &MM_COLS,
                                             &display_mm,
                                             &SMG$M_BORDER)) != SS$_NORMAL)
	panic(status,"create virtual display","init menu");
}

menu_options(list,title)
char **list[];
char *title;
{
    int status;
    int key;
    int i,y;

    if (!menu_inited)
        init_menu();

    if ((status = smg$label_border(&display_mm,
                                   DSC1(title),
                                   &SMG$K_TOP,
                                   0,
                                   &(SMG$M_REVERSE))) != SS$_NORMAL)
        panic(status,"label border displayf","menu options");

    status = smg$erase_display(&display_mm);
    if (status != SS$_NORMAL)
        panic(status,"erase display","menu options");

    for (i = 0; list[i] != 0; i++)
    {
        /*
        **  display this line on the display, starting at line MM_FIRST_LINE
        **  indent MM_INDENT
        */
        y = i + MM_FIRST_LINE;
    
        smg$set_cursor_abs(&display_mm,&y,&MM_INDENT);
        if ((status = smg$put_line(&display_mm,
                                   DSC1(list[i]))) != SS$_NORMAL)
            panic(status,"put line of menu option","menu options");
    }

    status = smg$paste_virtual_display(&display_mm,
                                       &paste,
                                       &MM_PASTE_Y,
                                       &MM_PASTE_X);
    if (status != SS$_NORMAL)
        panic(status,"paste virtual display","menu options");

    key = get_key(NO_AUTO);

    status = smg$unpaste_virtual_display(&display_mm,&paste);
    if (status != SS$_NORMAL)
       panic(status,"unpaste virtual display","menu options");

    if (key == SMG$K_TRM_CR)
        return(FALSE);

    return(key);
}

files_menu_options(file_type)
int file_type;
{
    /*
    **  if file_type == SHOW_FILES we're just showing non-dirs, else
    **  we're just showing dirs
    */
    if (file_type == SHOW_FILES)
        return(menu_options(fm_choices," FILE MANAGER OPTIONS "));
    else
        return(menu_options(dm_choices," FILE MANAGER (DIRECTORIES) OPTIONS "));
}

main_menu_options()
{
    return(menu_options(mm_choices," DIRECTORY MANAGER OPTIONS "));
}

define_menu_options()
{
    return(menu_options(df_choices," USER-DEFINABLE OPTIONS "));
}

stats_menu()
{
    return(menu_options(st_choices," DIRECTORY STATISTICS OPTIONS "));
}
