/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rms.h>
#include <ctype.h>
#include <ssdef.h>
#include <iodef.h>
#include <smgdef.h>
#include <smgmsg.h>
#include <smg$routines.h>
#include <fchdef.h>
#include <atrdef.h>
#include <fibdef.h>
#include <jpidef.h>
#include <starlet.h>
#include <lib$routines.h>
#include <uaidef.h>
#include <unistd.h>
#include <unixio.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_defs.h"

/*
**  maximum no of directory structures we'll try to load at startup
*/
#define MAX_ARGS	20

/*
**  acb elements
*/
ULONG uchar;

/*
**  acb descriptor for io$_access
*/
ACB acb[] =   {
               /*
               **  file characteristics (directory ??)
               */
               ATR$S_UCHAR,     ATR$C_UCHAR,     &uchar,

               /*
               **  terminate the list
               */
               0,               0,               0};

/*
**  blank fib, for initialising
*/
struct fibdef blank_fib;

/*
**  item list for getjpiw
*/
ITM_LST itmlst[2];

/*
**  save terminal characteristics
*/
char sense_mode[12];
char set_mode[12];

#ifdef __DECC

static void	catch_c(void);
static void	exit_routine(void);
static void	find_user(void);
static void	init(void);

#endif

/*
**  starting point for the program
*/
main(int argc, char *argv[])
{
    size_t c;
    ULONG    status;
    ULONG    uaflags;
    USHORT   i;
    IOSB_DEF iosb;
    char dir[SPEC_SIZE+1],name[SPEC_SIZE+1],tmp_dir[SPEC_SIZE+1];
    int arg_no, filmgr_start;
    globaldef int captive, try_000000, spec_to_root, comm_to_root;
    char arg_list[SPEC_SIZE+1];
    char arg[MAX_ARGS][SPEC_SIZE+1];
    int arg_decrement_flag;

    /* Small check to see if this is a supported terminal
    ** More info in swing_39.c
    */

    status = check_termtype();
    if(!status)
      { printf("Sorry, This is not a supported terminal type!\n");
        exit(0);}
    /*
    **  my version of sys$disk
    */
    translate("SYS$DISK",sys_disk);

    /*
    **  find out if we're captive
    */
    itmlst[0].buffer_length = 4;
    itmlst[0].item_code = JPI$_UAF_FLAGS;
    itmlst[0].buffer_address = &uaflags;
    itmlst[0].return_length_address = &i;
    itmlst[1].buffer_length = itmlst[1].item_code = 0;

    uaflags = 0;
    i = 0;
    status = sys$getjpiw(0,			/* efn */
                         0,			/* pidadr */
                         0,			/* prcnam */
                         &itmlst,		/* itmlst */
                         &iosb,			/* iosb */
                         0,			/* astadr */
                         0);			/* astprm */
    if((status & 1) != 1)
            panic(status,"getjpi","uaf flags");
    if((iosb.status & 1) != 1)
            panic(status,"getjpi","uaf flags");
    captive = (uaflags & UAI$M_CAPTIVE);

    /*
    **  check for qualifier args:
    **
    **  -h or /h  Show Help list of command line arguments
    **  -v or /v  Show Version of this program
    **  -f or /f  Start up in File_manager
    **  -0 or /0  Look for and use a device__000000__SWSVE.DAT file if a
    **                    device__directory__SWSVE.DAT file is not found
    **                    in the SWING$SAVE directory.
    **  -s or /s  Exits in SYS$SPECIFIC restore SYS$SYSROOT as default dir.
    **  -c or /c  Exits in SYS$COMMON   restore SYS$SYSROOT as default dir.
    */
    filmgr_start = FALSE;
#ifdef ALWAYS_TRY_000000
    try_000000 = TRUE;
#else
    try_000000 = FALSE;
#endif
#ifdef SPECIFIC_EXITS_TO_SYSROOT
    spec_to_root = TRUE;
#else
    spec_to_root = FALSE;
#endif
#ifdef COMMON_EXITS_TO_SYSROOT
    comm_to_root = TRUE;
#else
    comm_to_root = FALSE;
#endif

    if (argc > 1) {
	for (arg_no = 1; arg_no < argc; arg_no++)
	{
	    /*
	    **  test for qualifiers which evoke printing to screen and
	    **  then exit:  -h or /h, -v or /v
	    **
	    **  Note that -h will not output descriptions of qualifiers
	    **  other than -f if they have been made defaults via logicals
	    **  in SWING.H or via inclusion in CSWING's foreign command
	    **  definition.
	    */
	    arg_decrement_flag = FALSE;
	    if ((strcmp(argv[arg_no],"-h")==0) ||
	       (strncmp(argv[arg_no],"/h",2)==0)) {
		printf("\nUsage:\n");
		printf(" swing -h  Show this Help.\n");
		printf(" swing -v  Show Version of this program.\n");
		sprintf(junk," swing[ -f]");
		if (!try_000000)
		    strcat(junk,"[ -0]");
		if (!spec_to_root)
		    strcat(junk,"[ -s]");
		if (!comm_to_root)
		    strcat(junk,"[ -c]");
		strcat(junk,"[ dir][,dir...]\n\n");
		printf("%s",junk);
		printf("  -f  Start up directly in the File manager.\n");
		if (!try_000000) {
		 printf("  -0  Look for and use a device__000000__SWSVE.DAT\n");
		 printf("       file if a device__directory__SWSVE.DAT file\n");
		 printf("       is not found in the SWING$SAVE directory.\n");
		}
		if (!spec_to_root) {
		 printf("  -s  Exits in SYS$SPECIFIC restore SYS$SYSROOT as\n");
		 printf("       the default dir.\n");
		}
		if (!comm_to_root) {
		 printf("  -c  Exits in SYS$COMMON   restore SYS$SYSROOT as\n");
		 printf("       the default dir.\n");
		}
		printf(" dir  Use VMS syntax including brackets (e.g.,\n");
		printf("       [topdir.subdir] or device:[topdir.subdir]\n");
		printf("       or [.subdir]).  You can also use logical\n");
		printf("       dir names (e.g., SYS$EXAMPLES).  If a dir\n");
		printf("       is not specified, your current default is\n");
		printf("       assumed.\n\n");
		exit(1);
	    }
	    if ((strcmp(argv[arg_no],"-v")==0) ||
		(strncmp(argv[arg_no],"/v",2)==0)) {
		printf(VERSION);
		exit(1);
	    }

	    /*
	    **  test for qualifiers which modify program behavior: -f or /f
	    **  -0 or /0, -s or /s, -c or /c
	    */
	    if ((strcmp(argv[arg_no],"-f")==0) ||
	       (strncmp(argv[arg_no],"/f",2)==0))
	    {
		filmgr_start = TRUE;	/* Invoke file manager at start */
		for (i = arg_no; i+1 < argc; i++) argv[i] = argv[i+1];
		argc--;		/* Move args down and decrement arg count */
		arg_decrement_flag = TRUE;
	    }
	    if ((strcmp(argv[arg_no],"-0")==0) ||
	       (strncmp(argv[arg_no],"/0",2)==0))
	    {
		try_000000 = TRUE;	/* Try device root DAT file */
		for (i = arg_no; i+1 < argc; i++) argv[i] = argv[i+1];
		argc--;		/* Move args down and decrement arg count */
		arg_decrement_flag = TRUE;
	    }
	    if ((strcmp(argv[arg_no],"-s")==0) ||
	       (strncmp(argv[arg_no],"/s",2)==0))
	    {
		spec_to_root = TRUE; /* SYS$SPECIFIC->SYS$SYSROOT on exit */
		for (i = arg_no; i+1 < argc; i++) argv[i] = argv[i+1];
		argc--;		/* Move args down and decrement arg count */
		arg_decrement_flag = TRUE;
	    }
	    if ((strcmp(argv[arg_no],"-c")==0) ||
	       (strncmp(argv[arg_no],"/c",2)==0))
	    {
		comm_to_root = TRUE;    /* SYS$COMMON->SYS$SYSROOT on exit */
		for (i = arg_no; i+1 < argc; i++) argv[i] = argv[i+1];
		argc--;		/* Move args down and decrement arg count */
		arg_decrement_flag = TRUE;
	    }

	    /*
	    **  make sure we pick up "print_then_exit" qualifiers if they
	    **  were preceded in the command line by "modify_behavior"
	    **  qualifier(s)
	    */
	    if (arg_decrement_flag)
		arg_no -= 1;
	}
    }
		
    /*
    **  Establish start_dir as the device and directory from which program
    **  was invoked.  If Q_EXITS_TO_ORIGINAL_DIR is define in swing.h,
    **  entering Q or q in the Directory Manager causes program to exit 
    **  with start_dir as the default.
    */
    getcwd(start_dir,SPEC_SIZE+1);

    /*
    **  build up a list of the user-specified dirs in arg[][]
    */
    arg_list[0] = 0;
    arg_no = 0;
    if (argc > 1)
    {
        /*
        **  concatenate the startup directories
        */
        for (i = 1; i < argc; i++)
        {
            strcat(arg_list,argv[i]);
            strcat(arg_list,",");
        }

        /*
        **  build up a list of the startup directories
        */
        for (c = 0; c < strlen(arg_list) && arg_no < MAX_ARGS;)
        {
            strcpy(arg[arg_no],&arg_list[c]);

            for (i = 0; i < strlen(arg[arg_no]); i++)
            {
                if (arg[arg_no][i] == ',')
                {
                    arg[arg_no][i] = 0;
                    break;
                }
            }
            if (strlen(arg[arg_no]))
                arg_no++;

            c = c + i + 1;
        }
    }


    /*
    **  check we can set def to first dir - if not bomb
    **  on failure chdir returns non-zero
    */
    if (arg_no && chdir(arg[0]))
    {
        printf("Cannot change directory to %s\n",arg[0]);
        exit(1);
    }

    /*
    **  initialize
    */
    init();

    /*
    **  Main title
    */
    title(MTITLE);
    sub_title_underline(" ");
    /*
    **  get going - add a structure to the display - first current unless
    **  user specified others
    */    
    if (argc == 1)
    {
        show_def(dir);
        add_a_dstructure(dir);
    }
    else
    {
        for (i = 0; i < arg_no; i++)
        {
	    /*
	    **  if argument is a directory spec (i.e., without a device
	    **  spec, and not a logical), enable use of specs relative
	    **  to start_dir (e.g., [.junk] or [-.-.junk])
	    */
	    if (arg[i][0] == '[')
	    {
		show_def(dir);
		set_default(start_dir);    /* Yes, we need both of these */
		status = chdir(start_dir); /* calls for some situations. */
		status = chdir(arg[i]);
		if (status)
		{
		    sprintf(junk,
			    "Cannot access directory %c%s%c",
			    '"',
			    arg[i],
			    '"');
		    warning(junk);
		    set_default(dir);    /* Yes, we need both of these */
		    status = chdir(dir); /* calls for some situations. */
		}
		else
		{
		    show_def(dir);
		    if (!add_a_dstructure(dir))
		    {
			sprintf(junk,
				"Failed to find directory %c%s%c",
				'"',
				dir,
				'"');
                	warning(junk);
		    }
		}
	    }
	    /*
	    **  else argument has a device spec or is a logical,
	    **  so use it as is
	    */
	    else
	    {
		if (!add_a_dstructure(arg[i]))
		{
		sprintf(junk,
			"Failed to find directory %c%s%c",
			'"',
			arg[i],
			'"');
                warning(junk);
		}
	    }
        }
    }

    /*
    **  if we haven't got any dirs on display we'll add the sys$login
    **  dir just to make sure we have a dir on display
    **  this could happen if there was no dir specified and the current
    **  default is not valid
    **  this is a last-case precaution to ensure user gets started-up
    **  with a dir on display    
    */
    if (!fdp)
        add_a_dstructure("sys$login:");    

    /*
    **  solicit user input - effectively the body of the processing
    */
    next_command(TRUE,filmgr_start);

    /*
    **  exit
    */
    exit(exit_error = 1);
}

/*
**  find the root directory of the current directory structure -
**  this is the same as is shown by show def for example
**	from	where we start
**	dir	full dir spec
**	name	name of top dir
*/
void get_root_dir(char *from, char *dir, char *name)
{
    int status,s2,cntxt,start,size;
    char act[256],*ap,result[256];

    strcpy(act,from); ap=act;
    while (*ap) {
      if (*ap=='<') *ap='[';
      if (*ap=='>') *ap=']';
      ap++;
    }
    ap--;
    strcpy(dir,act); strcat(act,"*.*;*"); cntxt=0;
/*  printf(">>> Entry with %s <<<\n",act); /***/
    status=lib$find_file(DSC1(act),DSCA2(result),&cntxt,0,0,0,0);
    s2=lib$find_file_end(&cntxt);
    while (status&1||status==RMS$_FNF||status==RMS$_DNF) {
      act[strlen(act)-5]='\0'; strcpy(dir,act);
      while (*ap!='.' && *ap!='[') ap--;
      if (*ap=='.') {
        *ap++ =']'; *ap-- ='\0';
        strcat(act,"*.*;*"); cntxt=0;
/*      printf("Testing %s...\n",act); /***/
        status=lib$find_file(DSC1(act),DSCA2(result),&cntxt,0,0,0,0);
        s2=lib$find_file_end(&cntxt);
      } else status=0;
    }
/*  printf(">>> Exit with %s <<<\n",dir); getchar(); /***/

    ap=dir-1;
    while (*ap++) if (isalpha(*ap)) *ap=toupper(*ap);

    start=strcspn(dir,"[")+1;
    size=strcspn(&dir[start],"]");
    strncpy(name,&dir[start],size);
    name[size]=0;
}

/*
**  build up a list of the directories in the structure we're examining
*/
int get_dirs(NODE *n_ptr, ULONG channel)
{
    int i;
    int c = 0;
    int status;
    NODE *temp;
    struct dsc$descriptor_s fib_desc;
    struct fibdef fib;
    char filename[SPEC_SIZE+1];
    char dirname[SPEC_SIZE+1];
    short file_length;
    IOSB_DEF iosb;

    /*
    **  initialize
    */
    fib = blank_fib;

    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = (void *) &fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;

#ifdef __DECC
    fib.fib$w_did[0] = n_ptr->fid[0];
    fib.fib$w_did[1] = n_ptr->fid[1];
    fib.fib$w_did[2] = n_ptr->fid[2];
    fib.fib$w_nmctl = FIB$M_WILD;
#else
    fib.fib$r_did_overlay.fib$w_did[0] = n_ptr->fid[0];
    fib.fib$r_did_overlay.fib$w_did[1] = n_ptr->fid[1];
    fib.fib$r_did_overlay.fib$w_did[2] = n_ptr->fid[2];
    fib.fib$r_nmctl_overlay.fib$w_nmctl = FIB$M_WILD;
#endif

    filename[0] = 0;
    file_length = 0;
    while(1)
    {
        static char dir_template[20];
	strcpy(dir_template, "*.dir;*");
        uchar = 0;
        status = sys$qiow(0,
                          channel,
                          IO$_ACCESS,
                          &iosb,
                          0,
                          0,
                          &fib_desc,
                          DSC1(dir_template),
                          &file_length,
                          DSCA1(filename),
                          acb,
                          0);
        if((status & 1) != 1)
            panic(status,"io$_access directory lookup","get dirs");
        if(iosb.status == SS$_NOMOREFILES)
            return(TRUE);

        /*
        **  file in dir but not indexf.sys
        */
        if(iosb.status == SS$_NOSUCHFILE)
        {
/* removed this line since wcc is not supposed to be manipulated by progs */
/*            fib.fib$l_wcc++; */
            continue;
        }

        /*
        **  on error go get next file unless fib.l_wcc == 0; in which case
        **  we can't read the dir, so we bomb
        */
        if((iosb.status & 1) != 1)
        {
            if (fib.fib$l_wcc == 0)
            {
                /*
                **  we can't read this directory - go home
                */
                warning(full_error_text(iosb.status));
                return(iosb.status);
            }
            continue;
        }

        /*
        **  carry on
        */
        filename[file_length] = 0;

        /*
        **  if it was a directory (& not [000000].dir) we'll build it in
        */
        if (uchar & FCH$M_DIRECTORY &&
#ifdef __DECC
            (fib.fib$w_fid[0] != mfd_did[0] ||
             fib.fib$w_fid[1] != mfd_did[1]) )
#else
            (fib.fib$r_fid_overlay.fib$w_fid[0] != mfd_did[0] ||
             fib.fib$r_fid_overlay.fib$w_fid[1] != mfd_did[1]) )
#endif
        {
            /*
            **  we've a directory
            */
            c++;
            temp = n_ptr;

            /*
            **  we only want dirname to be filename, not extension;version
            */
            strcpy(dirname,filename);
            dirname[strcspn(dirname,".")] = 0;
            n_ptr = grab_node(dirname);

            /*
            **  record file id
            */
#ifdef __DECC
            n_ptr->fid[0] = fib.fib$w_fid[0];
            n_ptr->fid[1] = fib.fib$w_fid[1];
            n_ptr->fid[2] = fib.fib$w_fid[2];
#else
            n_ptr->fid[0] = fib.fib$r_fid_overlay.fib$w_fid[0];
            n_ptr->fid[1] = fib.fib$r_fid_overlay.fib$w_fid[1];
            n_ptr->fid[2] = fib.fib$r_fid_overlay.fib$w_fid[2];
#endif

            /*
            **  record directory id
            */
#ifdef __DECC
            n_ptr->did[0] = fib.fib$w_did[0];
            n_ptr->did[1] = fib.fib$w_did[1];
            n_ptr->did[2] = fib.fib$w_did[2];
#else
            n_ptr->did[0] = fib.fib$r_did_overlay.fib$w_did[0];
            n_ptr->did[1] = fib.fib$r_did_overlay.fib$w_did[1];
            n_ptr->did[2] = fib.fib$r_did_overlay.fib$w_did[2];
#endif

            if (c == 1)
            {
                /*
                **  I am the child of my parent
                */
                temp->child = n_ptr;
            }
            else
            {
                /*
                **  I am my sister's sister
                */
                temp->sister = n_ptr;
            }

            /*
            **  hunt more directories
            */
            get_dirs(n_ptr,channel);
        }
    }
}

/*
**  allocate space for a new directory structure
**	name	name of this directory
*/
NODE *grab_node(char *name)
{
    NODE *p;
    int i,len;
    
    if (p = calloc(1, sizeof (NODE)))
    {
        give_node_a_name(p,name);
        p->child = p->sister = p->lost = 0;
	if (!getenv("SWING_EXCLUDE_FILES"))
	{
	   strcpy(p->dir_spec,DEFAULT_DIR_SPEC);
	   p->in_or_ex_clude = INCLUDE_FILES;
	}
	else
	{
	   strcpy(p->dir_spec,getenv("SWING_EXCLUDE_FILES"));
	   p->in_or_ex_clude = EXCLUDE_FILES;
	}
        return(p);
    }

    panic(1,"cannot allocate any more memory","grab-node");
}

/*
**  update directory structure with directory name
*/
void give_node_a_name(NODE *p, char *name)
{
    int len, i;

    len = strlen(name);
    for (i = len-1; i >= 0; i--)
    {
        if (isprint(name[i]))
        {
            name[i+1] = 0;
            break;
        }
    }
    strcpy(p->name,name);
    new_short_name(p,name);
}

/*
**  change the short name
*/
void new_short_name(NODE *p, char *name)
{
    sprintf(p->short_name,
            "%-*.*s",
            NAME_SHORT_SIZE,
            NAME_SHORT_SIZE,
            name);
    if (strlen(name) > NAME_SHORT_SIZE)
        p->short_name[NAME_SHORT_SIZE-1] = '*';
}

/*
**  allocate space for a display line pointer
*/
LINE *grab_line(void)
{
    LINE *p;
    static LINE *last_grabbed_line;
    int i;

    if (p = calloc(1, sizeof (LINE)))
    {
        curr_l_ptr = p;

        if (total_lines)
        {
            /*
            **  if this is not the first line we've grabbed
            */
            p->back_ptr                 = last_grabbed_line;
            last_grabbed_line->forw_ptr = p;
        }
        else
        {
            first_line_ptr = p;
        }
        
        last_grabbed_line = p;
        total_lines++;
        for (i = 0; i < MAX_LEVEL ; i++)
        {
            p->connections[i] = 0;
            p->n_ptr[i] = 0;
        }
        return(p);
    }

    panic(1,"cannot allocate any more memory","grab_line");
}

void load_line(LINE *l_ptr, int line, int level, NODE *n_ptr,DSTRUCTURE *dp)
{
    LINE *temp;

    /*
    **  only tie up the first MAX_LEVEL nodes into the l_ptr structure
    **  (entries 0 to MAX_LEVEL-1) If there are more nodes, e.g. we're
    **  starting from [000000] we'll not tie up these extra nodes, but
    **  we'll set the last connection to show a diamond, e.g. there are
    **  nodes for which we ain't got room
    */

    l_ptr->dp = dp;
    l_ptr->n_ptr[level] = n_ptr;
    n_ptr->line = line;
    n_ptr->level = level;

    /*
    **  recover any 'lost' nodes
    */
    if (n_ptr->lost)
    {
        n_ptr->child = n_ptr->lost;
        n_ptr->lost = 0;
    }

    if (n_ptr->child)
    {
        /*
        **  if we've gone as far as we can go, i.e. we've filled in our
        **  right-most dir then we'll start to throw away
        */        
        if (level+1 == MAX_LEVEL)
        {
            /*
            **  Because we only hold MAX_LEVEL levels, we'll
            **  break off this part of the structure and return 
            **  the nodes under lost. We'll indicate there are lost nodes
            **  by leaving a DEC special graphics diamond behind.
            **  We can recover lost nodes later.
            */
            n_ptr->lost = n_ptr->child;
            n_ptr->child = 0;
            l_ptr->connections[level] = C_DIAMOND;
        }
        else
        {
            /*
            **  think of the child of this node
            */
            l_ptr->connections[level] = C_LEFT + C_RIGHT;
            load_line(l_ptr, line, level+1, n_ptr->child, dp);
        }
    }

    if (n_ptr->sister)
    {
        /*
        **  think of the sisters
        */
        temp = grab_line();
        load_line(temp, total_lines - 1, level, n_ptr->sister, dp);
        add_sister_bars(temp, level);
    }
}

/*
**  catch control c
*/
static void catch_c(void)
{
    /*
    **  exit if we're not doing so already - if we're exiting let it continue
    **  to exit
    */
    if (controlc_on == FALSE)
    {
        /*
        **  invoke exit handler if we should
        */
        if (!panicked_already)
            exit(1);
    }
}

/*
**  exit routine - linked into the system in init() - this will be called
**  when process terminates
*/
static void exit_routine(void)
{
    USHORT op_channel;
    ULONG  status;
    ULONG op_chlong;
    int rows,columns,i;
    char default_dir[SPEC_SIZE+1];
    DSTRUCTURE *tmp;
    globalref int spec_to_root, comm_to_root;

    /*
    **  switch off all future calls to panic
    */
    panicked_already = TRUE;

    /*
    **  Recreate pasteboards & displays
    **  They are destroyed by SMG exit handler
    */
    status = smg$create_pasteboard(&paste);

    rows    = DISPLAY_DCL_ROWS;
    columns = DISPLAY_DCL_COLUMNS;
    status = smg$create_virtual_display(&rows,
             &columns,
             &display_dcl);

    /*
    **  ensure cursor switched on
    */
    smg$set_cursor_mode(&paste,&0);

    /*
    **  if we were in pasteboard update we'll get out of it
    */
    smg$end_pasteboard_update(&paste);


    /*
    **  disable broadcast trapping
    */
    disable_broadcasts();

    /*
    **  if we've changed the directory structure in any way we'll save it
    */
    tmp = fdp;
    while(tmp)
    {
        if (tmp->to_be_saved)
            save_nodes(tmp,TRUE);
        tmp = tmp->forw_ptr;
    }

    /*
    **  make sure we clear the vdu by showing a blank display
    **  which we can write to if need be
    */
    smg$erase_display(&display_dcl);
    smg$paste_virtual_display(&display_dcl,&paste,&1,&1);

    /*
    **  ensure we're not in special grapics by shovelling out the necessary
    **  escape sequence to switch us to acsii - is this a hack?
    */
    printf("%c(B%c(B%c(B",ESCAPE,ESCAPE,ESCAPE);
    fflush(stdout);

    /*
    **  restore terminal width
    */
    i = 255 & sense_mode[2];
    smg$change_pbd_characteristics(&paste, &i, 0, 0, 0, 0, 0);

    if (broadcasts_rxed)
    {
        /*
        **  show the broadcast display if any broadcasts were received
        **  - nice feature if you don't like losing messages
        */
        smg$paste_virtual_display(&display_msg,&paste,&2,&2);
        smg$set_cursor_abs(&display_dcl,&21,&1);
    }
    else
        smg$set_cursor_abs(&display_dcl,&1,&1);

    /*
    **  reset numeric keypad if original state
    */
    if ((sense_mode[10] & 128) == 0) smg$set_keypad_mode(&keyboard,&0);

    /*
    **  get current default directory
    */
    show_def(default_dir);

    /*
    **  if SPECIFIC_EXITS_TO_SYSROOT is defined in SWING.H or -s or /s were
    **     used as command line arguments,
    **  and the default device is SYS$SPECIFIC,
    **  and sysroot_flag is set,
    **  then make the default device SYS$SYSROOT.
    */
    if (spec_to_root && strstr(default_dir,"SYS$SPECIFIC") && sysroot_flag) {
	strcpy(junk,"SYS$SYSROOT:");
	i = 0;
	while (default_dir[i++] != '[');
        strcat(junk,&default_dir[i-1]);
	strcpy(default_dir,junk);
	set_default(default_dir);
    }

    /*
    **  if COMMMON_EXITS_TO_SYSROOT is defined in SWING.H or -c or /c were
    **     used as command line arguments,
    **  and the default device is SYS$COMMON,
    **  and sysroot_flag is set,
    **  then make the default device SYS$SYSROOT.
    */
    if (comm_to_root && strstr(default_dir,"SYS$COMMON") && sysroot_flag) {
	strcpy(junk,"SYS$SYSROOT:");
	i = 0;
	while (default_dir[i++] != '[');
        strcat(junk,&default_dir[i-1]);
	strcpy(default_dir,junk);
	set_default(default_dir);
    }

    /*
    **  show user where we are
    */
    sprintf(junk,"Current directory is %s",default_dir);
    smg$put_line(&display_dcl,DSC1(junk),0,0,0,&1);

    /*
    **  restore ctrl-c, ctrl-t to state they were at the beginning of the
    **  program
    */
    enable_ctrl();

    /*
    **  get a channel to the op device
    */
    status = sys$assign(DSC1("SYS$OUTPUT"),&op_channel,0,0);
    if((status & 1) != 1)
    {
        panic(status, "error assigning op_channel","init");
    }

    /*
    **  reset terminal to entry state
    */
    op_chlong = op_channel;
    op_chlong = (op_chlong & 65535);
    status = sys$qiow(0,
                 op_chlong,
                 IO$_SETMODE,
                 0,
                 0,
                 0,
                 sense_mode,
                 12,
                 0,
                 0,
                 0,
                 0);
    if((status & 1) != 1)
    {
        panic(status,"error setting mode","init");
    }

    puts(" ");

    status = sys$dassgn(op_channel);
    if((status & 1) != 1)
        panic(i,"sys$dassgn","init");
    
    /*
    **  if error show error via VMS
    */
    exit(exit_error);
}

/*
**  enable ctrl
*/
void enable_ctrl(void)
{
    controlc_on = TRUE;
    lib$enable_ctrl(&old_msk);
}

void disable_ctrl(void)
{
    /*
    **  switch off ctrl-c, ctrl-y, ctrl-t, make a note in old_mask of the
    **  initial settings (for exit_routine)
    */
    ULONG  status;
    int mask;

    controlc_on = FALSE;
    mask = CTRLT + CTRLY;
    status = lib$disable_ctrl(&mask,&old_msk);
    if((status & 1) != 1)
    {
        panic(status,"error disabling ctrl_t and ctrl_y","disable ctrl");
    }
}

/*
**  initialise swing
*/
static void init(void)
{
    ULONG  status;
    USHORT op_channel;
    ULONG op_chlong;
    int i,rows,columns,esc_flag,attributes;
/*    char sense_mode[12]; moved to beginning of module */

    /*
    **  for $DCLEXH
    */
    static struct _exit_block {
    int forward;
    void  *address;
    int zero;
    void  *condition;
    } exit_handler_block;

    static ULONG condition;

    /*
    **  get a channel to sys$input terminal
    */
    status = sys$assign(DSC1("SYS$INPUT"),&channel1,0,0);
    if((status & 1) != 1)
    {
        panic(status,"error assigning channel","init");
    }

    /*
    **  get a channel to the op device
    */
    status = sys$assign(DSC1("SYS$OUTPUT"),&op_channel,0,0);
    if((status & 1) != 1)
    {
        panic(status,"error assigning op_channel","init");
    }

    /*
    **  check we have at least PAGE_MIN_LENGTH rows
    */
    op_chlong = op_channel;
    op_chlong = (op_chlong & 65535);
    status = sys$qiow(0,
                 op_chlong,
                 IO$_SENSEMODE,
                 0,
                 0,
                 0,
                 sense_mode,
                 12,
                 0,
                 0,
                 0,
                 0);
    if((status & 1) != 1)
    {
        panic(status,"error sensing mode","init");
    }

    for (i = 0; i < 12; i++) set_mode[i] = sense_mode[i];
    if ((set_mode[4] & 16) == 0)	/* no hostsynch... turn it on */
    {
	set_mode[4] = set_mode[4] | 16;
        op_chlong = op_channel;
        op_chlong = (op_chlong & 65535);
	status = sys$qiow(0,
                 op_chlong,
                 IO$_SETMODE,
                 0,
                 0,
                 0,
                 set_mode,
                 12,
                 0,
                 0,
                 0,
                 0);
        if((status & 1) != 1)
	{
	    panic(status,"error setting mode","init");
	}
    }
    /*
    **  check we have at least PAGE_MIN_LENGTH rows
    */

    page_length = (int)sense_mode[7];
    if (page_length < PAGE_MIN_LENGTH)
    {
        printf("Page length is only %d lines\n",page_length);
        printf("Must be at least %d lines\n",PAGE_MIN_LENGTH);
        exit(1);
    }

    puts(" ");

    status = sys$dassgn(op_channel);
    if((status & 1) != 1)
        panic(status,"sys$dassgn","init");
    
    /*
    **  switch off ctrl-c, ctrl-y, ctrl-t
    */
    disable_ctrl();

    /*
    **  original location of exit handler declaration was here
    */

    /*
    **  find out who the user is, load into this_user
    */
    find_user();

    /*
    **  wind up SMGs - pasteboard
    */
    status = smg$create_pasteboard(&paste);
    if((status & 1) != 1)
        panic(status,"create pasteboard","init_smugs");

    status = smg$erase_pasteboard(&paste);
    if((status & 1) != 1)
        panic(status,"erase pasteboard","init_smugs");

    /*
    **  keyboard
    */
    status = smg$create_virtual_keyboard(&keyboard);
    if((status & 1) != 1)
        panic(status,"create virtual keyboard","init_smugs");

    modify_pbd();

    /*
    **  declare an exit handler
    **
    **    Code moved to this location by WBF on 8/3/90.
    **    In VMS V5.2+, exit handler must be declared AFTER calls to
    **    smg$create_pasteboard and smg$create_virtual_keyboard routines.
    **    See VMS V5.2 Release Notes, section 4.31.5.3.
    */
    exit_handler_block.forward = 0;
    exit_handler_block.address = (void* )&exit_routine;
    exit_handler_block.zero    = 0;
    exit_handler_block.condition = &condition;

    status = sys$dclexh(&exit_handler_block);
    if((status & 1) != 1)
    {
        panic(status,"error declaring an exit handler","init");
    }

    /*
    **  top line display
    */
    rows    = DISPLAYT_ROWS;
    columns = DISPLAYT_COLUMNS;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &displayt,
                                   0,
                                   0);
    if((status & 1) != 1)
        panic(status,"create virtual display displayt","init_smugs");

    /*
    **  now paste - this is the first display pasted -
    **  if we can't support SMG$ then we'll get the error SMG$_WILUSERMS - 
    **  so we say to the user he should get a better terminal
    */
    rows    = DISPLAYT_Y;
    columns = DISPLAYT_X;
    status = smg$paste_virtual_display(&displayt, &paste, &rows, &columns);
    if(status == SMG$_WILUSERMS)
    {
        strcpy(junk,getenv("TERM"));
        for (i = 0; i < strlen(junk); i++)
            junk[i] = toupper(junk[i]);

	printf("ERROR - this terminal does not support the SMG$ routines\n");
        printf("      - this current terminal type is [%s]\n",junk);
        printf("      - a DEC vt terminal or workstation is required\n");
        printf("      - SWING exiting\n");
        exit(1);
    }

    if((status & 1) != 1)
        panic(status,"pasting display t","init_smugs");

    /*
    **  display for the directories
    */
    rows    = DISPLAY1_ROWS;
    columns = DISPLAY1_COLUMNS;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &display1);
    if((status & 1) != 1)
        panic(status,"create virtual display display1","init_smugs");

    rows    = DISPLAY1_Y;
    columns = DISPLAY1_X;
    status = smg$paste_virtual_display(&display1, &paste, &rows, &columns);
    if((status & 1) != 1)
        panic(status,"pasting display 1","init_smugs");

    esc_flag = 1;
    status = smg$allow_escape(&display1, &esc_flag);
    if((status & 1) != 1)
        panic(status,"allow escape","init_smugs");

    /*
    **  bottom line display
    */
    rows    = DISPLAYB_ROWS;
    columns = DISPLAYB_COLUMNS;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &displayb,
                                   0,
                                   0);
    if((status & 1) != 1)
        panic(status,"create virtual display displayb","init_smugs");

    rows    = DISPLAYB_Y;
    columns = DISPLAYB_X;
    status = smg$paste_virtual_display(&displayb, &paste, &rows, &columns);
    if((status & 1) != 1)
        panic(status,"pasting display b","init_smugs");

    /*
    **  dcl display
    */
    rows    = DISPLAY_DCL_ROWS;
    columns = DISPLAY_DCL_COLUMNS;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &display_dcl);
    if((status & 1) != 1)
        panic(status,"create virtual display display_dcl","init_smugs");

    /*
    **  files display
    */
    rows    = DISPLAYF_ROWS;
    columns = DISPLAYF_COLUMNS;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &displayf,
                                   &SMG$M_BORDER);
    if((status & 1) != 1)
        panic(status,"create virtual display displayf","init_smugs");

    /*
    **  received broadcast messages
    */
    rows = MESSAGE_ROWS;
    columns = MESSAGE_COLS;
    attributes = SMG$M_BORDER;
    status = smg$create_virtual_display(&rows,
                                   &columns,
                                   &display_msg,
                                   &attributes /*,
                                   &SMG$M_BOLD */);
    if((status & 1) != 1)
        panic(status,"create virtual display display_msg","init");

    sprintf(junk," BROADCAST MESSAGES for %s ",this_user);
    status = smg$label_border(&display_msg,DSC1(junk),
                         &SMG$K_TOP,0,&(SMG$M_REVERSE));
    if((status & 1) != 1)
        panic(status,"label border display_msg","init");

    /*
    **  catch ctrl-c, ctrl-y then we're finished
    */
    status = sys$qiow(0,channel1,IO$_SETMODE + IO$M_CTRLCAST + IO$M_CTRLYAST,
                 0,0,0,&catch_c,0,0,0,0,0);
    if((status & 1) != 1)
    {
        panic(status,"error catching ctrl c","init");
    }

    /*
    **  once we've enabled broadcasts
    */
    enable_broadcasts();

    /*
    **  load user's definitions
    */
    load_definitions(TRUE);
}


/*
**  work out where to position the cursor when we start up -
**  do this by 1) finding the directory structure we're at, then
**  2) finding where this structure is on the display
*/
void set_init_position(void)
{
    USHORT fid[3];
    LINE *ptr;
    NODE *n;
    char dir[SPEC_SIZE+1];
    char device[SPEC_SIZE+1];
    int i;

    /*
    **  get fid of current directory
    */
    show_def(dir);
    strcpy(device,dir);
    get_did(dir,fid);

    /*
    **  strip dirs off from device
    */
    for (i = 0; i < strlen(device); i++)
        if (device[i] == '[')
        {
            device[i] = 0;
            break;
        }

    ptr = first_line_ptr;

    for (current_line = 0; current_line < total_lines; current_line++)
    {
        for (current_level = 0; current_level < MAX_LEVEL; current_level++)
        {
            n = ptr->n_ptr[current_level];
            if (n && fid[0] == n->fid[0] &&
                     fid[1] == n->fid[1] &&
                     fid[2] == n->fid[2] &&
                     strcmp(device,ptr->dp->device) == 0)
			goto label_1;
        }
        ptr = ptr->forw_ptr;
    }
    
    /*
    **  not found, so leave at top
    */
    current_level = current_line = 0;
    sprintf(junk,"directory %s not found",dir);
    warning(junk);
    set_default(curr_l_ptr->dp->dir_name);
    search_nodes(curr_l_ptr->dp->dir_name,
                 curr_l_ptr->dp->name,
                 curr_l_ptr->dp);
    load_lines();
    set_default(dir);

label_1:
    /*
    **  If we haven't started the display we'll try to put the current 
    **  selection in the middle row of the screen. If we have started then we'll
    **  not move the top and bottom lines of the screen unless we have to. 
    **  If we have to we'll try to put the current selection in the middle 
    **  row of the screen.
    */
    if (!started || 
        current_line < current_top_line ||
        current_line > (current_top_line + DISPLAY1_ROWS - 1))
    {
        current_top_line = current_line - DISPLAY1_ROWS/2;
    }

    /*
    **  if current_top_line too great ...
    */
    if (current_top_line + DISPLAY1_ROWS > total_lines)
    {
        current_top_line = total_lines - DISPLAY1_ROWS;
    }

    /*
    **  if negative (too small)
    */
    if (current_top_line < 0)
    {
        current_top_line = 0;
    }

    /*
    **  set curr_l_ptr to point to line current_line
    */
    curr_l_ptr = current_line_ptr(current_line);
}

/*
**  return a LINE type pointer to the structure containing the info for the
**  n'th line of the display
*/
LINE *current_line_ptr(int line)
{
    int i;
    LINE *temp;

    temp = first_line_ptr;

    for (i = 0; i < line; i++)
        temp = temp->forw_ptr;

    return(temp);
}

/*
**  search the directory structure, winkling out all the directories
*/
int search_nodes(char *dir, char *name, DSTRUCTURE *dp)
{
    ULONG  status;
    USHORT channel;
    USHORT fid[3];

    /*
    **  get fid of starting directory
    */
    if (!get_did(dir,fid))
        return(FALSE);

    /*
    **  assign channel to current device
    */
    status = sys$assign(DSC1("SYS$DISK"),&channel,0,0);
    if((status & 1) != 1)
        panic(status,"sys$assign","search nodes");

    sprintf(junk,"Searching, root = %s ",dir);
    message(junk);

    /*
    **  free off the existing nodes
    */
    free_nodes(dp->fnp);
    dp->fnp = grab_node(name);
    dp->fnp->fid[0] = fid[0];
    dp->fnp->fid[1] = fid[1];
    dp->fnp->fid[2] = fid[2];

    /*
    **  don't know my parent fid, but it will not be used anyway
    */
    dp->fnp->did[0] = 0;
    dp->fnp->did[1] = 0;
    dp->fnp->did[2] = 0;

    get_dirs(dp->fnp,channel);

    dp->to_be_saved = TRUE;
    save_nodes(dp,TRUE);

    status = sys$dassgn(channel);
    if((status & 1) != 1)
        panic(status,"sys$dassgn","search nodes");

    return(TRUE);
}

/*
**  make sure we're in 80 column mode
*/
void modify_pbd(void)
{
    ULONG  status;
    int width = SCREEN_WIDTH;

    status = smg$change_pbd_characteristics(&paste, &width, &original_width,
                                       0, 0, 0, 0);
    if((status & 1) != 1)
        panic(status,"change pbd chars","modify pbd");

}

/*
**  free the pointed to line structure and all other line structures
**  that it references
*/
void free_lines(LINE *lptr)
{
    /*
    **  free this line and all lines it points to
    */
    LINE *temp, *tmp;

    tmp = lptr;
    while(tmp)
    {

        temp = tmp->forw_ptr;
        free(tmp);
        tmp = temp;
    }
}

/*
**  free the directory - node structure, children and sisters
*/
int free_nodes(NODE *n_ptr)
{
    /*
    **  free nodes, all sisters and children
    */
    NODE *c, *s, *l;

    if (!n_ptr)
        return(TRUE);

    c = n_ptr->child;
    s = n_ptr->sister;
    l = n_ptr->lost;

    /*
    **  remove from list of visited dirs if it's in the list
    */
    remove_visited_dir(n_ptr);

    free(n_ptr);

    if (c)
        free_nodes(c);

    if (s)
        free_nodes(s);

    if (l)
        free_nodes(l);
}

/*
**  load up this_user as [GROUP,USER]
**  if we can't translate the uic, we'll just set this_user up as
**  getenv("USER")
*/
static void find_user(void)
{
    unsigned int group;
    int status;
    int namlen;
    char nambuf[41];

    group = (getgid() << 16) + 0xffff;

    namlen = 0;
    status = sys$idtoasc(group,
                         &namlen,
                         DSCA1(nambuf),
			 0,
			 0,
			 0);
    if((status & 1) != 1)
    {
	nambuf[namlen] = 0;
	sprintf(this_user,
		"[%s,%s]",
		nambuf,
		getenv("USER"));
    }
    else
	strcpy(this_user,getenv("USER"));
}

    /*
    Some bugfix have been made to correct the following Bugs (Or features).

 900803 Foteos Macrides  (MACRIDES@WFEB2.BITNET)
	In VMS V5.2+, an exit handler must be declared after calls to
	smg$create_pasteboard and smg$create_virtual_keyboard routines 
	or the results will be unpredicatable.  

 901006 (MRL@NERUS.PFC.MIT.EDU)
	Fix the pasteboard is destroyed by the SMG exit handler.  
	The screen isn't erased upon exiting (and isn't set back 
	to width=80 if originally was 132)
	Allow SWING to search and see if a stored data structure 
	is present for DISK:[000000] rather than stopping the 
	search at the top level directory.
	Fix iterative translations for logical names that did 
	not include search lists, translate only if the logical 
	name would eventually translate to a logical name that 
	contained a search list.
	Fix problems if a colon was not present.

 901212 (petechen@cs.rutgers.edu)
	Added some abilty to use Logical, se localfix.c and localfix.h

 910504 Harry Flowers (FLOWERS@MSUVX1.MEMST.EDU)
      o Did not handle disks in bound volume sets.
      o Couldn't exit from "@ - change file protection".
      o Disk quota ("&") option did not stay on screen until key pressed.
      o Quit option did not work on copy or move files or directory options.
      o More ("Y") option failed after invoking it a fixed number of times.
      o More ("Y") option ^W did not reset internally to normal from wide.
      o More ("Y") option searching did not work in case-insensitive mode
	unless a lower-case string was used for the search pattern.
      o More ("Y") option spacing problem with form feeds; now treated just
	like a non-printable character (replaces it with a "^" character).
      o More ("Y") option caused the file manager to display the file in
	lower case after it had been viewed.

 910507 Mats Akerberg (mats@efd.lth.se)
	Added the extra MATS_HACK becuse i like CR to EXIT and some
	don't, now you just define MATS_HACK in SWING.H.
	Fixed problem with long lines in configuration files.

 910613 Harry Flowers (flowers@memstvx1.memst.edu)
      o /file_manager or -f command line option
	This option specified on the command line will act just as if "F"
	were entered as soon as you get the directory manager screen.
      o Old problem with change protection only changing world protection
	has been fixed.
      o Terminal characteristics (like numeric keypad) are now restored on
	exiting from CSWING.

 910613 Bernd Onasch (ONASCH@ira.uka.de)
	Modified routine get_root_dir from SWING.C.  I updated this routine
	because I think normally no unpriv'd user has (or should have :-)
	access to the root directory of his/her directory tree.

 910729 Carl Karcher (KARCHER@WAISMAN.WISC.EDU)
	Modified SWING_12.C to ignore the non-displayable broadcast messages
	sent to VWS terminal windows.  Otherwise whenever the VWS window
	running swing is selected, the broadcast paste buffer would pop up
	and try to display the message and fail with a SMG$_NONBRDMSG
	status.

 910729 Charles Bailey (BAILEY@hmivax.humgen.upenn.edu)
      o Patch to SWING_24.C causes add_dstructure() to respect concealed
	and terminal attributes of device LNMs.
      o Patches to SWING_8.C, SWING_35.C move screen refresh after file
	manager @,~ out to command parser, and leave pointer at current
	file after command executes.

 910829 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Fixed bug with negative percentages on large file copies
      o Preserve file creation date on copy like DCL copy does
      o Fixed memory leak; variables were being trashed when the starting
	directory was more than 40 characters in length; caused many bad
	side-effects

 910831 Tony McGrath (TMcG@Physics.Monash.Edu.Au)
	Fixed bug when Swing is started up on an empty directory the
	program said that the SWSVE.DAT was out of date, coz the routine
	to find the ROOT dir of the current structure had failed.

 910913 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Change to volume set patch to handle very large file ids
      o Added file exclude "#" option and SWING_EXCLUDE_FILES logical
	to work like DIR/EXC, but include "=" option is exclusive; also
	file purge would have to have been extensively re-worked, so
	it's not supported while files are excluded

 910922 Mats Akerberg (mats@efd.lth.se)
      o Added a check for supported terminaltype 

 920415 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Upgraded MOST to latest version with many bug fixes.

 920427 Holger Teutsch (teutsch@ffm-iw.Degussa.DE)
      o Fixed $sndjbc items used in submitting jobs for VMS V5.5.

 920618 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Changed behavior when Swing is invoked in an unknown subdirectory
	in a known structure.  It used to just place you back at the top
	and give an unknown directory error.  Now, it rebuilds the dir
	structure and places you where you wanted to start.

 920626 Foteos Macrides (MACRIDES@WFEB2.BITNET)
      o Made code for Q_EXITS_TO_ORIGINAL_DIR more efficient.
      o Fixed bugs when getting the version limits of files for display
	with format_5.  If directory files are listed, their default
	version limits (for files created within the directories) are
	also displayed.
      o Modified to skip defragmentation of any files which have a version
	limit set (otherwise, such files would be trashed; the proper fix
	would be to eliminate the version limit, defragment the file, then
	restore its version limit, but since many sites now run defrag
	software this doesn't seem worth the bother).

 920701 Mark London (MRL@NERUS.PFC.MIT.EDU)
      o Fixed a bug in the code I submitted a while ago for searching for
	directories under a 000000 directory.  It doesn't work for
	directories that have a _ in their name.

 920921 Foteos Macrides (MACRIDES@WFEB2.BITNET)
      o Fixed more bugs in code for searching for directories under a
	000000 directory if the device__000000_SWSVE.DAT file exists
	and the device__directory__SWSVE.DAT file doesn't.  Made this
	behavior an option via a -0 or /0 qualifier.  It can be made
	the default (i.e., regardless of whether -0 or /0 is entered on
	the command line) by defining ALWAYS_TRY_000000 in SWING.H.
      o Enhanced or debugged aspects of logical device name handling.
	However, earlier modifications to respect terminal and concealed
	device logicals have broken some of CSWING's capacity to deal with
	search lists, and it can't all be restored without a major rewrite.
      o Added code such that arguments or logicals which translate to
	SYS$SYSROOT:[directory] will cause SYS$SPECIFIC:[directory] and
	SYS$COMMON:[directory] trees to be displayed (SYS$SPECIFIC is not
	in the SYS$SYSROOT search list, so it must be treated as a special
	case).  Optional defines (SPECIFIC_EXITS_TO_SYSROOT and
	COMMON_EXITS_TO_SYSROOT) in SWING.H enable setting the default
	device to SYS$SYSROOT on exit in SYS$SPECIFIC or SYS$COMMON if an
	original call was for SYS$SYSROOT.  Added a -s or /s and a -c or
	/c qualifier to evoke this behavior if SPECIFIC_EXITS_TO_SYSROOT
	and COMMON_EXITS_TO_SYSROOT, respectively, are not defined in
	SWING.H.  For all other search lists that can be handled, the
	default directory on exit is always that of the particular tree
	you were in at the time of exit.
      o Modified the command line (-h) help module to describe the -0, -s
	and -c qualifiers if they were not made defaults via logicals in
	SWING.H or by including them in the foreign command definition
	for CSWING.
      o Added capacity to handle relative directory specs (e.g., [.foo] or
	[-.-.foo]).  An argument with such a spec is handled relative to
	the directory from which CSWING was called, regardless of its
	position in an argument list (e.g., the list  SYS$MANAGER,[]  will
	put up trees for SYS$SPECIFIC:[SYSMGR], SYS$COMMON:[SYSMGR], and
	your calling default, in that order).
      o Added pause and prompt features for displays of CSWING.HLB (like
	those when reading help libraries via DCL).  The help page length
	is determined by the value of page_length, determined at startup
	of CSWING.

 921010 Foteos Macrides (MACRIDES@WFEB2.BITNET)
      o Modified swing_8.c and swing_38.c to enable calling MOST with a
	set of selected files, and thereby use MOST's windowing features
	for multiple files (and appropriately modified the quick help list
	in swing_22.c).  Made numerous modifications/enhancements of MOST
	for maximum compatibility with CSWING, as described in the header of
	MOST's main.c.  Most noteably, command keys are now the same as or 
	homologous to those of CSWING wherever possible.  MOST called from
	CSWING will read and use MOST_SWITCHES and MOST_EDITOR if defined,
	can spawn editors and will use that defined for CSWING if its own
	logical is not defined, and can spawn to DCL in a manner analogous
	to spawns from CSWING (or will respect the 'captive' restriction if
	set).  Also modified main.c so that a standalone MOST can be built
	with the same code at that for CSWING.  Added confirmation of intent
	to edit in swing_16.c and MOST (our users too often hit E with
	intent to exit, not edit).  Provided a MOST.DOC_FM for this version
	of MOST (Note that the original copyright applies to it, and to any
	programs derived from it:  You must freely distribute the source
	code.).

 921128 Foteos Macrides (MACRIDES@SCI.WFEB.EDU, MACRIDES@WFEB2.BITNET)
      o Upgraded MOST to 3.03FM, as described in the header of its main.c.

 930106 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Changed file view option when files are selected to prompt for
	viewing the current file if selected files are not viewed.

 930111 Harry Flowers (Flowers@Memstvx1.MemSt.Edu)
      o Fixed bug in showing broadcast messages; sometimes messages
	would fail to display, though the time banner would.  This was
	caused by an incorrect declaration of the message length, which
	would sometimes be corrupted as a result.

 930129 Foteos Macrides (MACRIDES@SCI.WFEB.EDU, MACRIDES@WFEB2.BITNET)
      o Upgraded MOST to 3.04FM.  Added SecureMode extend switch (+s or +S)
	for use with gopher client, and other fixes/enhancements described
	in the header of MOST's main.c.

 930305 Foteos Macrides (MACRIDES@SCI.WFEB.EDU, MACRIDES@WFEB2.BITNET)
	Major version upgrade, to CSWING v3.7:
      o Upgraded MOST to 3.2.0FM.  Added lots of new features and switches
	as described in the header of its main.c and in most.doc_fm.  Most
	notably, now has both Quick Help (summary of commands) and Full Help
	(display of doc file), GOLD and EDT style commands, -h qualifier for
	showing Full Help from the DCL command line, -n Naive Prompt
	qualifer and logical for replacing the dev:[dir]filename string in
	the status line with other information or a title.
      o Ported everything to DECC/Alpha, with tests for VAXC versus DECC
	in the code and build procedures.  Thanks to Bill Forbes,
	BFORBES@MR4DEC.ENET.DEC.COM, for help with VAXC versus DECC
	differences.  Thanks to Ross Smith, SMITH@MCCLB0.MED.NYU.EDU, for
	letting me use NYU Med's field test Alpha.
      o Updated the SWING Help Library and SWING Manual.

 930315 Harry Flowers (Flowers@MSUVX2.MemSt.Edu)
      o Fixed captive account editor selection (where there is no user
	profile) to not allow selection of "Other" editor commands, as
	any DCL command could be entered.

 930715 Butch Walker (Butch-Walker@uokhsc.edu)
      o (Enabled by _PRINT_MENU) The addition of a print menu if a ? is
	typed for the queue name.  The menu includes those queues which
	have a description field formatted as specified in PRINTMENU.README
      o The Print command can now print the current file, rather than
	requiring that the user select a file.
      o (Changes to SWING.MMS) Added a DEBUG macro which compiles and links
	with debug and places the .OBJ files in a [.dbg] directory. This
	allows switching between debug and nodebug versions with minimal
	recompilation.  Added a USESCA macro which creates an SCA library
	in [.scalib], compiles with /ana and loads the .ANA modules into
	the library.  This allows users with SCA and LSE to make use of
	the utilities for maximum effectiveness.

 930715 Harry Flowers (Flowers@MSUVX2.MemSt.Edu)
      o Added debug changes to command files; the BUILD command file will
	now accept "DEBUG" as a parameter and work with either MMS or the
	command files.

 930720 Nissim Etrog (ETROG1@vms.huji.ac.il)
      o Better handling of iosb status block in swing_29.c (deals more
	intelligently with WARNING or INFO versus ERROR or SEVERE).
      o Enabled passing of parameters to execute_DCL().
      o Added informational messages for presses of undefined command keys.

 930720 Foteos Macrides (Macrides@sci.wfeb.edu)
      o Changed headers of the MATS_HACK and CR_RETURNS_FROM_FILE_TO_DIR_MGR
	logicals in swing.h to make more clear what they do and how they
	interact, and modified the quick help menu for the Directory manager
	along the lines submitted by Jeff Dykzeul (jd@mass.dnet.hac.com) so
	that it's correct, based on how they were defined.  Also got the
	quick help menu to be correct no matter what combinations of these
	two logicals and the Q_EXITS_TO_ORIGINAL_DIR logical are defined.
      o Sleep for only one second after warning messages to bottom of screen.

 931015 Foteos Macrides (Macrides@sci.wfeb.edu)
      o Improved MOST's spawning behavior and updated its doc file.
      o Updated C Swing manual and help library.

 931021 Richard Hardy (hardy@spike.wcc.govt.nz)
      o Fixed problem with Listing a file, and specifing the queue as a batch
        queue, allowed a captive user to submit a command procedure.

 931109 Harry Flowers (Flowers@MSUVX2.MemSt.Edu)
      o Changed default behavior of [Return] in the file manager from a
	no-op to the same as "Y" to view files.  This means that [Return]
	always has a function in the file manager; it either views the
	file(s) or returns to the directory manager depending on whether
	or not CR_RETURNS_FROM_FILE_TO_DIR_MGR is defined.

 931119 Harry Flowers (Flowers@Narnia.MemSt.Edu)
      o Replaced SWING-STARTUP.COM with a totally new SWING_STARTUP.COM

 931123 Harry Flowers (Flowers@Narnia.MemSt.Edu)
      o Performance enhancements using multiblock and multibuffer counts
	to reduce I/O's.

 940404 Harry Flowers (Flowers@Narnia.Memphis.Edu)
      o Fixed problem with purge... it wouldn't purge large numbers of
	previous versions of files.
      o Fixed problem with displaying renamed files; if the checksum
	happend to match, it wouldn't be updated.  Swaping parts of the
	name on 4-byte boundaries at the correct offset always results
	in the same checksum, so the display wouldn't be updated.  Yes,
	someone actually ran into this problem, and I had a heck of a
	time reproducing the problem and tracking it down!
      o Some versions of DEC C have trouble finding fchdef.h... so, I
	define DECC$TEXT_LIBRARY to be SYS$LIBRARY:SYS$LIB_C.TLB for
	them (yuch!).  Of course, the fix may break when they fix the
	bug. :-(

 940701 Glenn Everhart
      o The logical CSWING$USERCOMMAND can be predefined as a setup option.
	If it is defined, then the command G (or g) in the file manager
	will cause a scratch file to be written and to have all selected
	filenames written to it, one per line with newline between, and
	the user command defined in the logical will be executed via a
	system() call with the filename of the scratch file passed as an
	argument on the command line. Thus it can be picked up by a DCL
	script or by a program.  This is provided so that CSwing can be
	used as a file selector front end for pretty much anything you like,
	without particularly touching the CSwing program itself.  After the
	user application is done, the scratch file is deleted.

 940803 Jesper Partoft (syntese@login.dkuug.dk)
      o Fixed bug in get_did with no return status on success that caused
	problems in later versions of DEC C.

 941119 Harry Flowers (Flowers@Narnia.Memphis.Edu)
      o Several folks reported problems not having hostsync on, so now
	C Swing turns it on if it is off; it's restored to its original
	setting on exit.  Probably, these folks should have it on all
	the time for other programs as well, but we won't worry about
	that here. ;-)  This was typically noticed when repeating arrow
	keys would return A, B, C, and D (depending on the key pressed).
      o Cleanup of option "G" code for user defined commands.
      o Added code to keep someone from moving a directory under itself.
      o Changed check from a logical to the compiler image for DEC C so
	that it will test positive on a VAX with DEC C.
*/
