/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_20 - a real file copy function using RMS
*/
#include <stdio.h>
#include <string.h>
#include <ssdef.h>
#include <rms.h>
#include <starlet.h>
#include <lib$routines.h>
#include <str$routines.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

/*
**  transfer up to 32kb at a time (64 x 512 byte blocks)
**  this makes for faster copy, as we're copying on same device, so we'll
**  get less head movement and a better response
**
**  if we're going over decnet use a smaller buffer so that we see the 
**  transfer more clearly
*/

#define BUFFER_SIZE_BIG		32768
#define BUFFER_SIZE_SMALL	4096

#define DISPLAY_WIDTH  50

#define CREATE	"creating -"
#define COPY    "         - copying -"
#define CLOSE	"                   - closing"

ULONG lib_copy_file(char *from, char *to, char *filename, 
                    int remaining, int option, int prot)
{
    struct FAB fab1, fab2;
    struct RAB rab1, rab2;
    struct XABFHC xab1, xab2;
    struct XABPRO xabpro1, xabpro2;
    struct XABDAT xabdat1, xabdat2;
    char file_buffer[BUFFER_SIZE_BIG];
    int status, buffer_size;
    int index, sub_index;
    unsigned long total_bytes, bytes_copied = 0;
    char copy_line[SCREEN_WIDTH+1];
    char device[SPEC_SIZE+1];

    fab1 = cc$rms_fab;
    fab2 = cc$rms_fab;
    rab1 = cc$rms_rab;
    rab2 = cc$rms_rab;
    xab1 = cc$rms_xabfhc;
    xab2 = cc$rms_xabfhc;
    xabpro1 = cc$rms_xabpro;
    xabpro2 = cc$rms_xabpro;
    xabdat1 = cc$rms_xabdat;
    xabdat2 = cc$rms_xabdat;

    /*
    **  show what's happening
    */
    strcpy(device,to);
    device[strcspn(device,"[")] = 0;

    /*
    **  if there's any node info (::) we'll set the buffer size smaller
    **  for display purposes
    */
    buffer_size = BUFFER_SIZE_BIG;

    status = str$find_first_substring(DSC1(from),
                                      &index,
                                      &sub_index,
                                      DSC2("::"));
    if((status & 1) == 1)
        buffer_size = BUFFER_SIZE_SMALL;

    status = str$find_first_substring(DSC1(to),
                                      &index,
                                      &sub_index,
                                      DSC2("::"));
    if((status & 1) == 1)
    {
        buffer_size = BUFFER_SIZE_SMALL;

	/*
        **  don't show quotas for far away devices
        */
        device[0] = 0;
    }

    /*
    **  now update the copy window
    */
    copy_window(remaining, filename, device);

    /*
    **  set up fab1, rab1 (input file)
    */
    fab1.fab$b_fac = FAB$M_BIO | FAB$M_GET;
    fab1.fab$l_fna = from;
    fab1.fab$b_fns = strlen(from);
    fab1.fab$l_xab = &xab1;
    rab1.rab$l_bkt = 0;
    rab1.rab$l_fab = &fab1;
    rab1.rab$l_ubf = file_buffer;
    rab1.rab$w_usz = buffer_size;
    xab1.xab$l_nxt = &xabdat1;
    if (prot)
        xabdat1.xab$l_nxt = &xabpro1;

    /*
    **  open input file for block mode i/o
    */
    copy_window_percent(CREATE,0,1);

    status = sys$open(&fab1);
    if (status != RMS$_SUC)
        return(status);

    status = sys$connect(&rab1);
    if (status != RMS$_SUC)
    {
        sys$close(&fab1);
        return(status);
    }

    /*
    **  copy all FAB atributes from input file to output file
    */
    lib$movc3(&FAB$C_BLN,	/* length of a FAB */
              &fab1,		/* source = input file FAB */
              &fab2);		/* destn  = output file FAB */

    /*
    **  copy all XAB atributes from input file to output file
    */
    lib$movc3(&XAB$C_FHCLEN,	/* length of XABFHC */
              &xab1,		/* source = input file XAB */
              &xab2);           /* destn  = output file XAB */

    lib$movc3(&XAB$C_PROLEN,	/* length of XABPRO */
              &xabpro1,		/* source = input file XAB */
              &xabpro2);        /* destn  = output file XAB */

    /*
    **  set up fab2, rab2, xab2 (output file)
    */
    fab2.fab$w_ifi = 0;			/* clear internal file id */
    fab2.fab$b_fac = FAB$M_BIO | FAB$M_PUT;
    fab2.fab$l_fna = to;
    fab2.fab$b_fns = strlen(to);
    fab2.fab$l_fop |= FAB$M_MXV;	/* maximise ver no of created file */
    fab2.fab$l_xab = &xab2;
    rab2.rab$l_bkt = 0;
    rab2.rab$l_fab = &fab2;
    rab2.rab$l_rbf = file_buffer;
    xab2.xab$w_verlimit = 0; 		/* no limit on no of versions */
    xab2.xab$l_nxt = &xabdat2;
    xabdat2.xab$q_cdt = xabdat1.xab$q_cdt;	/* preserve creation date */

    if (prot)
    {
        xabdat2.xab$l_nxt = &xabpro2;
	xabpro2.xab$w_pro = ~prot;
    }

    /*
    **  size of file in bytes
    */
    total_bytes = (unsigned) fab1.fab$l_alq * 512;	

    /*
    **  now try to create the new output file
    */

    status = sys$create(&fab2);
    if (status != RMS$_SUC)
    {
        sys$close(&fab1);
        return(status);
    }

    status = sys$connect(&rab2);
    if (status != RMS$_SUC)
    {
        sys$close(&fab1);
        fab2.fab$l_fop = FAB$M_DLT;
        sys$close(&fab2);
        return(status);
    }

    if (option == EXPORT || option == IMPORT)
    {
        copy_window_percent(COPY,0,1);
    }

    /*
    **  copy loop - block by block
    */
    while(1)
    {
        status = sys$read(&rab1);
        if (status == RMS$_EOF)
        {
            /*
            **  it looks nice if we claim to have copied all the file - 
            **  in fact we might not copy all the file, if it's sequential
            **  and we've got to eof (for example)
            **
            **  so if we're not showing 100% show it
            */
            copy_window_percent(COPY,1,1);
            break;
        }
        else if (status != RMS$_SUC)
        {
            /*
            **  uh - error
            */
            sys$close(&fab1);
            fab2.fab$l_fop = FAB$M_DLT;
            sys$close(&fab2);
            return(status);
        }
        else
        {
            rab2.rab$w_rsz = rab1.rab$w_rsz;
            status = sys$write(&rab2);
            if (status != RMS$_SUC)
            {
                /*
                **  uh - error
                */
                sys$close(&fab1);
                fab2.fab$l_fop = FAB$M_DLT;
                sys$close(&fab2);
                return(status);
            }

            bytes_copied += (unsigned) rab1.rab$w_rsz;
            if (option == IMPORT || option == EXPORT)
            {
                label_copy_window(device);
                copy_window_percent(COPY,bytes_copied,total_bytes);
            }
        }
    }

    /*
    **  the end of successful operation
    */
    label_copy_window(device);
    copy_window_percent(CLOSE,1,1);

    sys$close(&fab1);
    sys$close(&fab2);

    return(SS$_NORMAL);
}
