/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_21 - a simple file locator
*/
#include <stdio.h>
#include <string.h>
#include <iodef.h>
#include <rms.h>
#include <ssdef.h>
#include <smgdef.h>
#include <smgmsg.h>
#include <fibdef.h>
#include <starlet.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

#define SEE_DIRECTORY		0
#define KEEP_ON_SEARCHING	1
#define NO_MORE_SEARCH          2
#define STAY_HERE		3

static char name_str[SPEC_SIZE+1];

/*
**  a blank fib used for initialising other fibs
*/
struct fibdef blank_fib;

static ULONG search_for_file(NODE *nptr, char *name, int level, 
                             char *found_file);
static ULONG search_for_the_file(char *device, NODE *nptr, char *name, 
                                 char *found_file);


/*
**  find a file, any file, get user to say which one
*/
int locate_file(void)
{
    int my_line, my_level;
    NODE *my_nptr;
    LINE *my_lptr;
    char found_file[SPEC_SIZE+1];
    char exp_str[SPEC_SIZE+1];
    char path[NAME_SIZE+1];
    int status, i;
    struct FAB fab;
    struct NAM nam;


    /*
    **  we'll prompt user with last file he searched for
    */
    strcpy(path,name_str);
    if (!get_user_input("File to locate [NONE] ",path,NAME_SIZE))
        return(FALSE);

    /*
    **  now let's just check that this string is a kosher file spec
    */

    fab = cc$rms_fab;
    nam = cc$rms_nam;

    exp_str[0] = 0;

    fab.fab$l_dna = "*.*;*";
    fab.fab$b_dns = 5;
    fab.fab$b_bid = FAB$C_BID;
    fab.fab$b_bln = FAB$C_BLN;
    fab.fab$l_fop = FAB$M_NAM;
    fab.fab$l_nam = &nam;
    fab.fab$l_fna = path;
    fab.fab$b_fns = strlen(path);

    nam.nam$b_nop = NAM$M_SYNCHK;
    nam.nam$b_bid = NAM$C_BID;
    nam.nam$b_bln = NAM$C_BLN;
    nam.nam$l_rsa = 0;
    nam.nam$b_rss = 0;
    nam.nam$b_rsl = 0;
    nam.nam$l_esa = exp_str;
    nam.nam$b_ess = SPEC_SIZE;

    status = sys$parse(&fab);
    if((status & 1) != 1)
    {
        sprintf(junk,"%s - Invalid file specification",
                error_text(status));
        warning(junk);
        return(FALSE);
    }
    sprintf(name_str,"%.*s",nam.nam$b_name+nam.nam$b_type+nam.nam$b_ver,
                     nam.nam$l_name);
    sprintf(junk,"Locate %s Y/N [N]",name_str);
    if (!confirm_user(junk,"Y"))
        return(FALSE);

    sprintf(junk,"Locating %s ...",name_str);
    message(junk);

    my_line = current_line;
    my_level = current_level;

    my_lptr = current_line_ptr(my_line);
    my_nptr = my_lptr->n_ptr[my_level];

    render(current_line,current_level,NORMAL_RENDITION);

    i = search_for_file(my_nptr,name_str,0,found_file);
    if (i != STAY_HERE)
    {
        /*
        **  well, we didn't find it
        */
        current_line = my_line;
        current_level = my_level;
        curr_l_ptr = current_line_ptr(current_line);
        redo_display(FALSE);

        if (i == KEEP_ON_SEARCHING)
        {
            sprintf(junk,"Failed to locate %s ...",name_str);
            warning(junk);
        }
    }
    render(current_line,current_level,SPECIAL_RENDITION);
    message("");
}

static ULONG search_for_file(NODE *nptr, char *name, int level, char *found_file)
{
    int i;

    current_line = nptr->line;
    current_level = nptr->level;
    curr_l_ptr = current_line_ptr(current_line);
    redo_display(FALSE);
    render(nptr->line,nptr->level,SPECIAL_RENDITION);

    /*
    **  set default in case we do a ^C
    **  if we can't we'll assume the user will want to keep on searching
    **  the sisters of the node (obviously can't search the children)
    */
    if (set_default(nptr->dir_name))
    {
        /*
        **  now search in this directory for the file
        */
        i = search_for_the_file(curr_l_ptr->dp->device,
                                nptr,
                                name,
                                found_file);
    
        render(nptr->line,nptr->level,NORMAL_RENDITION);    
    
        if (i != KEEP_ON_SEARCHING)
            return(i);
    
        /*
        **  file not found
        */
        if (!nptr->child && !nptr->sister)
            return(KEEP_ON_SEARCHING);
    
        if (nptr->child)
        {
            i = search_for_file(nptr->child,name,1,found_file);
            if (i != KEEP_ON_SEARCHING)
                return(i);
        }
    }
    else
        render(nptr->line,nptr->level,NORMAL_RENDITION);    

    /*
    **  we don't look for sisters of the first node that we search
    */
    if (level && nptr->sister)
    {
        i = search_for_file(nptr->sister,name,1,found_file);
        if (i != KEEP_ON_SEARCHING)
            return(i);
    }

    return(KEEP_ON_SEARCHING);
}

static ULONG search_for_the_file(char *device, NODE *nptr, char *name, 
                                 char *found_file)
{
    /*
    ** o.k. - does dir did in device contain a file called name
    */
    ULONG  status;
    USHORT channel;
    IOSB_DEF iosb;
    int i,file_length,orig_line,orig_level;
    char filename[SPEC_SIZE+1];
    char out_file[SPEC_SIZE+1];
    char dir_spec[NAME_SIZE+1];
    short in_or_ex_clude;
    char search_spec[NAME_SIZE+1];

    struct dsc$descriptor_s fib_desc;
    struct fibdef fib;

    /*
    **  initialize
    **
    **  when we do a dir we do it on a name.type basis, not nam.type;version
    **  so find the ; in name and sling it
    */
    strcpy(search_spec,name);
    search_spec[strcspn(search_spec,";")] = 0;

    /*
    **  assign channel to current device
    */
    channel = 0;
    status = sys$assign(DSC1(device),
                        &channel,
                        0,
                        0);
    if((status & 1) != 1)
        panic(status,"sys$asign","search for the file");

    fib = blank_fib;

    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = (void *) &fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;

#ifdef __DECC
    fib.fib$w_did[0] = nptr->fid[0];
    fib.fib$w_did[1] = nptr->fid[1];
    fib.fib$w_did[2] = nptr->fid[2];
    fib.fib$w_nmctl = FIB$M_WILD;
#else
    fib.fib$r_did_overlay.fib$w_did[0] = nptr->fid[0];
    fib.fib$r_did_overlay.fib$w_did[1] = nptr->fid[1];
    fib.fib$r_did_overlay.fib$w_did[2] = nptr->fid[2];
    fib.fib$r_nmctl_overlay.fib$w_nmctl = FIB$M_WILD;
#endif

    filename[0] = 0;
    file_length = 0;
    while(1)
    {
        status = sys$qiow(0,
                          channel,
                          IO$_ACCESS,
                          &iosb,
                          0,
                          0,
                          &fib_desc,
                          DSC1(name),
                          &file_length,
                          DSCA1(filename),
                          0,
                          0);
        if((status & 1) != 1)
            panic(status,"io$_access directory lookup","search for the file");
        if(iosb.status == SS$_NOMOREFILES)
        {
            sys$dassgn(channel);
            return(KEEP_ON_SEARCHING);
        }

        /*
        **  file in dir but not indexf.sys
        */
        if (iosb.status == SS$_NOSUCHFILE)
        {
/*            fib.fib$l_wcc++; */
            continue;
        }

        /*
        **  on error go get next file unless fib.l_wcc == 0; in which case
        **  we can't read the dir, so we bomb
        */
        if((iosb.status & 1) != 1)
        {
            if (fib.fib$l_wcc == 0)
            {
                /*
                **  we can't read this directory - go home
                */
                message(full_error_text(iosb.status));
                sys$dassgn(channel);
                return(KEEP_ON_SEARCHING);
            }

            /*
            **  carry on searching
            */
            continue;
        }
        
        message("");

        filename[file_length] = 0;

        while(1)
        {
            sprintf(junk,"Found %s  What now ? [See directory, Next file, Remain here, Quit]",
                    filename);
            message(junk);
            switch(get_key(NO_AUTO))
            {
case SMG$K_TRM_LOWERCASE_S:
case SMG$K_TRM_UPPERCASE_S:
                current_line = nptr->line;
                current_level = nptr->level;
		orig_line = current_line;
		orig_level = current_level;
                curr_l_ptr = current_line_ptr(current_line);
                set_default(nptr->dir_name);
		strcpy(dir_spec,nptr->dir_spec);
		in_or_ex_clude = nptr->in_or_ex_clude;
                strcpy(nptr->dir_spec,search_spec);
		nptr->in_or_ex_clude = INCLUDE_FILES;
                swing_files(nptr,filename,SHOW_FILES);

		/*
		**  look to see if we left the file manager in a different
		**  directory than we started, if so restore as was
		*/
		if (current_level != orig_level ||
		    current_line != orig_line)
		{
		    curr_l_ptr = current_line_ptr(orig_line);
                    nptr = curr_l_ptr->n_ptr[orig_level];
		    set_default(nptr->dir_name);
		    set_init_position();
                    refresh_display();
		}
                strcpy(nptr->dir_spec,dir_spec);
		nptr->in_or_ex_clude = in_or_ex_clude;
                break;

case SMG$K_TRM_LOWERCASE_R:
case SMG$K_TRM_UPPERCASE_R:
                sys$dassgn(channel);
                return(STAY_HERE);
                break;

case SMG$K_TRM_LOWERCASE_N:
case SMG$K_TRM_UPPERCASE_N:
                sprintf(junk,"Locating %s ...",name);
                message(junk);
                goto out_of_while;
                break;

case SMG$K_TRM_LOWERCASE_Q:
case SMG$K_TRM_UPPERCASE_Q:
                sys$dassgn(channel);
                return(NO_MORE_SEARCH);
                break;

default:
                warning("Invalid Input ...");
            }
        }

        /*
        **  go find next file
        */
out_of_while:;
    }
}
