/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_35 - a simple file protection changer
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <iodef.h>
#include <ssdef.h>
#include <smgdef.h>
#include <smgmsg.h>
#include <fibdef.h>
#include <starlet.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

#define SYSTEM_OFFSET	0
#define OWNER_OFFSET	4
#define GROUP_OFFSET	8
#define WORLD_OFFSET	12

#define READ_OFFSET	0
#define WRITE_OFFSET	1
#define EXEC_OFFSET	2
#define DEL_OFFSET	3

static int  determine_prot(char *string, USHORT *protection);
static int  get_owner_from_user(ULONG *uic);
static int  get_prot_from_user(USHORT *prot);
static void new_string(char *string);
static int  protect_a_dir(NODE *nptr, USHORT prot, int option, ULONG new_uic);
static int  protect_sub_dirs(NODE *nptr, USHORT prot, int level, int option,
                             ULONG new_uic);


/*
**  a blank fib used for initialising other fibs
*/
struct fibdef blank_fib;
static failed_to_change;

/*
**  protect all files in these directories
**
**  option == TRUE ==> protection
*/
void protect_dirs(int option)
{
    int my_line, my_level;
    NODE *my_nptr;
    LINE *my_lptr;
    char found_file[SPEC_SIZE+1];
    char exp_str[SPEC_SIZE+1];
    char path[NAME_SIZE+1];
    char string[SPEC_SIZE+1];
    unsigned short prot;
    ULONG  new_uic;
    int status, i;

    failed_to_change = 0;

    my_line = current_line;
    my_level = current_level;

    my_lptr = current_line_ptr(my_line);
    my_nptr = my_lptr->n_ptr[my_level];

    show_deleting(my_nptr,DELETING_RENDITION,FALSE);

    if (option)
        i = confirm_user("Change protection on these directories Y/N [N] ",
			 "Y");
    else
        i = confirm_user("Change ownership of these directories Y/N [N] ",
			 "Y");

    if (!i)
    {
        show_deleting(my_nptr,NORMAL_RENDITION,FALSE);
	goto prot_over;
    }

    if (option)
        i = get_prot_from_user(&prot);
    else
	i = get_owner_from_user(&new_uic);

    if (!i)
    {
        show_deleting(my_nptr,NORMAL_RENDITION,FALSE);
	goto prot_over;
    }

    show_deleting(my_nptr,NORMAL_RENDITION,FALSE);

    protect_sub_dirs(my_nptr,prot,0,option,new_uic);

    current_line = my_line;
    current_level = my_level;
    curr_l_ptr = current_line_ptr(current_line);
    redo_display(FALSE);

prot_over:;

    render(current_line,current_level,SPECIAL_RENDITION);

    if (failed_to_change)
    {
        sprintf(junk,"Failed to change %d files",failed_to_change);
	warning(junk);
    }
    else
    {
	sleep(1);
        message("");
    }
}

static int protect_a_dir(NODE *nptr, USHORT prot, int option, ULONG new_uic)
{
    /*
    ** o.k. - protect every file herein
    */
    ULONG  status;
    USHORT channel;
    IOSB_DEF iosb;
    int file_length;
    char filename[SPEC_SIZE+1];
    unsigned short dummy;
    struct dsc$descriptor_s fib_desc;
    struct fibdef fib;

    /*
    **  initialize

    /*
    **  assign channel to current device
    */
    channel = 0;
    status = sys$assign(DSC1("SYS$DISK"),
                        &channel,
                        0,
                        0);
    if((status & 1) != 1)
        panic(status,"sys$asign","protect a dir");

    fib = blank_fib;

    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = (void *) &fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;

#ifdef __DECC
    fib.fib$w_did[0] = nptr->fid[0];
    fib.fib$w_did[1] = nptr->fid[1];
    fib.fib$w_did[2] = nptr->fid[2];
    fib.fib$w_nmctl = FIB$M_WILD;
#else
    fib.fib$r_did_overlay.fib$w_did[0] = nptr->fid[0];
    fib.fib$r_did_overlay.fib$w_did[1] = nptr->fid[1];
    fib.fib$r_did_overlay.fib$w_did[2] = nptr->fid[2];
    fib.fib$r_nmctl_overlay.fib$w_nmctl = FIB$M_WILD;
#endif

    filename[0] = 0;
    file_length = 0;
    while(1)
    {
        status = sys$qiow(0,
                          channel,
                          IO$_ACCESS,
                          &iosb,
                          0,
                          0,
                          &fib_desc,
                          DSC1("*.*;*"),
                          &file_length,
                          DSCA1(filename),
                          0,
                          0);
        if((status & 1) != 1)
            panic(status,"io$_access directory lookup","protect a dir");
        if (iosb.status == SS$_NOMOREFILES)
        {
            sys$dassgn(channel);
            return(TRUE);
        }

        /*
        **  file in dir but not indexf.sys
        */
        if (iosb.status == SS$_NOSUCHFILE)
        {
/*            fib.fib$l_wcc++; */
            continue;
        }

        /*
        **  on error go get next file unless fib.l_wcc == 0; in which case
        **  we can't read the dir, so we bomb
        */
        if((iosb.status & 1) != 1)
        {
            if (fib.fib$l_wcc == 0)
            {
                /*
                **  we can't read this directory - go home
                */
                message(full_error_text(iosb.status));
                sys$dassgn(channel);
                return(TRUE);
            }

            /*
            **  carry on searching
            */
            continue;
        }
        
        filename[file_length] = 0;
        if (option)
            sprintf(junk,"changing protection on %s",filename);
	else
            sprintf(junk,"changing ownership on %s",filename);

	message(junk);

        /*
        **  go find next file
        */
        if (option)
	{
            status = change_the_protection(channel,
#ifdef __DECC
                                      fib.fib$w_fid,
#else
                                      fib.fib$r_fid_overlay.fib$w_fid,
#endif
                                      prot, &dummy);
            if((status & 1) != 1)
	        failed_to_change++;
	}
	else
	{
	    status = change_the_ownership(channel,
#ifdef __DECC
                                      fib.fib$w_fid,
#else
                                      fib.fib$r_fid_overlay.fib$w_fid,
#endif
                                      new_uic);
            if((status & 1) != 1)
	        failed_to_change++;
	}
    }
}

static int protect_sub_dirs(NODE *nptr, USHORT prot, int level, int option, ULONG new_uic)
{
    int i;

    current_line = nptr->line;
    current_level = nptr->level;
    curr_l_ptr = current_line_ptr(current_line);
    redo_display(FALSE);

    /*
    **  first of all the files in this directory
    */
    render(nptr->line,nptr->level,SPECIAL_RENDITION);
    protect_a_dir(nptr,prot,option,new_uic);
    render(nptr->line,nptr->level,NORMAL_RENDITION);    
    
    /*
    **  no more directories
    */
    if (!nptr->child && !nptr->sister)
        return(TRUE);
    
    if (nptr->child)
        protect_sub_dirs(nptr->child,prot,level+1,option,new_uic);

    /*
    **  we don't look for sisters of the first node that we search
    */
    if (level && nptr->sister)
        protect_sub_dirs(nptr->sister,prot,level,option,new_uic);

    return(TRUE);
}

int protect_files(NODE *n, int file_type, int option)
{
    FILER *temp;

    ULONG  status;
    USHORT prot, dummy;
    ULONG  new_uic;

    if (!files_selected)
    {
        message("No files selected");
        sleep(1);
        return(FALSE);
    }
    else 
    {
	if (option)
	{
            if (files_selected == 1)
                sprintf(junk,
                        "change protection on the selected file Y/N [N] ? ");
            else
                sprintf(junk,
                        "change protection on the %d selected files Y/N [N] ? ",
                        files_selected);
 	}
	else
	{
            if (files_selected == 1)
                sprintf(junk,
                        "change ownership of the selected file Y/N [N] ? ");
            else
                sprintf(junk,
                        "change ownership of the %d selected files Y/N [N] ? ",
                        files_selected);
 	}

        if (!confirm_user(junk,"Y"))
           return(FALSE);
    }

    if (option)
    {
        if (!get_prot_from_user(&prot))
    	    return(FALSE);
    }
    else
    {
	if (!get_owner_from_user(&new_uic))
	    return(FALSE);
    }

    temp = first_file_ptr;
    failed_to_change = 0;

    while(temp)
    {
        if (temp->rendition)
        {
            if (option)
	    {
                /*
                **  change protection on this file
                */
                sprintf(junk,
                        "Changing protection on %s",
                        temp->filename);
	        message(junk);
                status = change_the_protection(dir_channel,
                                          temp->fid, prot, &dummy);
                if((status & 1) != 1)
		    failed_to_change++;
            }
	else
	    {
                /*
                **  change ownership of this file
                */
                sprintf(junk,
                        "Changing ownership of %s",
                        temp->filename);
	        message(junk);
                status = change_the_ownership(dir_channel,
                                        temp->fid, new_uic);
                if((status & 1) != 1)
		    failed_to_change++;
            }
	}
        temp = temp->forw_ptr;
    }

    if (failed_to_change)
    {
        sprintf(junk,"Failed to change %d files",failed_to_change);
	warning(junk);
    }
    else
    {
	sleep(1);
    }

    n->checksum = 0;

    return(TRUE);
}

static int get_prot_from_user(USHORT *prot)
{
    char string[SCREEN_WIDTH+1];
    unsigned short tmp_prot;

    /*
    **  get user's desired input - first set string up with
    **  current default protection
    */
    sys$setdfprot(0, &tmp_prot);
    while(1)
    {
        prot_to_string(tmp_prot,string);
	strcat(string,"? ");
        if (!get_user_input_with_prompt("New file protection: ",
	                                string,
	                                SCREEN_WIDTH))
	    return(FALSE);

        if (!determine_prot(string,&tmp_prot))
        {
	    warning("Invalid input");
	    return(FALSE);
        }
        prot_to_string(tmp_prot,string);
        *prot = tmp_prot;
        sprintf(junk,"Really change protection to %s Y/N [N] ?",string);
        if (confirm_user(junk,"Y"))
	    return(TRUE);
	else
	    return(FALSE);
    }
}

static int determine_prot(char *string, USHORT *protection)
{
    int i, offset;
    unsigned short prot;
    prot = 0;

    new_string(string);

    i = 0;
    while(string[i])
    {
        switch(string[i++])
	{
case 'W':   offset = WORLD_OFFSET;
	    break;

case 'G':   offset = GROUP_OFFSET;
	    break;

case 'O':   offset = OWNER_OFFSET;
	    break;

case 'S':   offset = SYSTEM_OFFSET;
	    break;

default:
	    return(FALSE);

        }

	prot &= (-1 ^ (1 << (offset + READ_OFFSET)));
	prot &= (-1 ^ (1 << (offset + WRITE_OFFSET)));
	prot &= (-1 ^ (1 << (offset + EXEC_OFFSET)));
	prot &= (-1 ^ (1 << (offset + DEL_OFFSET)));

	while(string[i] &&
	      string[i] != ':' &&
              string[i] != ',')
	{
	    i++;
	}
        if (string[i] == ':')
	    i++;

	while(string[i] &&
	      string[i] != ',')
	{
	    switch(string[i++])
	    {
    case 'R':   prot |= 1 << (offset + READ_OFFSET);
		break;

    case 'W':   prot |= 1 << (offset + WRITE_OFFSET);
		break;

    case 'E':   prot |= 1 << (offset + EXEC_OFFSET);
		break;

    case 'D':   prot |= 1 << (offset + DEL_OFFSET);
		break;
	
    default:
		break;
	    }
	}
        if (string[i] == ',')
	    i++;
    }
    *protection = ~prot;
    return(TRUE);
}

static void new_string(char *string)
{
    int i,j,l;
    char s[2];

    j = 0;
    l = strlen(string);
    s[1] = 0;

    for (i = 0; i < l; i++)
    {
	s[0] = toupper(string[i]);
	if (s[0] == '?') s[0] = ',';
        if (strpbrk(":,WREDOGS",s))
	    string[j++] = s[0];
    }
    string[j] = 0;
}

static int get_owner_from_user(ULONG *uic)
{
    char user[SCREEN_WIDTH+1];
    int length, status, i;

    while(1)
    {
        if (!get_user_input("New file owner ? [NO CHANGE] ",
                            user,
                            SCREEN_WIDTH))
	    return(FALSE);

	/*
	**  check he / she / it exists
	*/
	status = sys$asctoid(DSC1(user),	/* name */
			     uic,		/* id */
			     0);		/* attrib */
        if((status & 1) != 1)
	{
	    warning(full_error_text(status));
	    continue;
	}

	for (i = 0; i < strlen(user); i++)
	    user[i] = toupper(user[i]);

	sprintf(junk,
		"Change ownership to %s - [%o,%o] Y/N [N] ?",
		user,
		((*uic & 0xffff0000) >> 16),
		(*uic & 0x0000ffff));

	if (confirm_user(junk,"Y"))
	    return(TRUE);
    }
}
