/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/

/*
**  swing_38
**
**  type/most routines
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <ssdef.h>
#include <smgdef.h>
#include <smgmsg.h>
#include <smg$routines.h>
#include <trmdef.h>
#include <unistd.h>
#include <lib$routines.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"
#include <descrip.h>

static void do_most_files(int num_most_args, char *most_args[]);


/*
**  type the current file or selected files via most
*/
int most_files(NODE *n)
{
    FILER *tmp;
    char *most_args[500];
    int num_most_args = 0;
    int screen;

    /*
    **  initial set of most_args[] could receive most's switches (-bcklstvw)
    **  and extended_switches (+d, +linenumber, +/string),
    **  so these lines are here as an example,
    **  but we don't use them.
    **
    most_args[num_most_args] = (char *) malloc(strlen("-csv")+1);
    strcpy(most_args[num_most_args++],"-csv");
    most_args[num_most_args] = (char *) malloc(strlen("+d")+1);
    strcpy(most_args[num_most_args++],"+d");
    */

    if (!files_selected)
    {
        /*
        **  most the current file
        */
        if (access(curr_file_ptr->filename,0) == 0)
        {
            tmp = current_file_ptr(curr_file_no);
            most_args[num_most_args] = (char *) malloc(strlen(tmp->filename)+1);
            strcpy(most_args[num_most_args++],tmp->filename);
        }
        else
        {
            sprintf(junk,"%s is not accessable",curr_file_ptr->filename);
            warning(junk);
            return(FALSE);
        }
    }
    else
    {
        /*
        **  confirm intention to most the selected files
        */
        if (files_selected == 1)
            sprintf(junk,"type the selected file via most Y/N [N]");
        else
            sprintf(junk,"type the %d selected files via most Y/N [N]",
                    files_selected);
        if (!confirm_user(junk,"Y"))
	{
	    sprintf(junk,"type the current file via most Y/N [N]");
	    if (!confirm_user(junk,"Y")) return(FALSE);
	    /*
	    **  most the current file
	    */
	    if (access(curr_file_ptr->filename,0) == 0)
	    {
		tmp = current_file_ptr(curr_file_no);
		most_args[num_most_args] = (char *) malloc(strlen(tmp->filename)+1);
		strcpy(most_args[num_most_args++],tmp->filename);
	    }
	    else
	    {
		sprintf(junk,"%s is not accessable",curr_file_ptr->filename);
		warning(junk);
		return(FALSE);
	    }
	}
	else
	{
        /*
        **  most the selected files
        */
        tmp = first_file_ptr;
        while(tmp)
        {
            if (tmp->rendition)
            {
                /*
                **  yup, this one's been chosen
                */
                if (access(tmp->filename,0) == 0)
                {
                most_args[num_most_args] =
                    (char *) malloc(strlen(tmp->filename)+1);
                strcpy(most_args[num_most_args++],tmp->filename);
                }
                else
                {
                sprintf(junk,"%s is not accessable",tmp->filename);
                warning(junk);
                }
            }
            tmp = tmp->forw_ptr;
        }
	}
    }

    /*
    **  if no files were accessible, return
    **  else do it
    */
    if (!num_most_args)
        return(FALSE);
    else
    {
        smg$save_physical_screen(&paste,&screen);
        disable_broadcasts();

            do_most_files(num_most_args,most_args);

        enable_broadcasts();
        smg$restore_physical_screen(&paste,&screen);
    }

    return(TRUE);
}

ULONG most_exception(int *sigarr, int *mecharr)
{
    /*
    **  unwind and return
    */
    lib$sig_to_ret(sigarr, mecharr);
    return(SS$_UNWIND);
}

static void do_most_files(int num_most_args, char *most_args[])
{
    VAXC$ESTABLISH(most_exception);
    most_less(num_most_args,most_args);
}

/*
**  Make a listfile of selected filespecs, then pass its spec to
**  a user defined command in a subprocess. Should permit Cswing to be
**  used as a frontend for lots of interesting stuff, with no loss
**  or prejudice to the rest of the program. The user command should be
**  preconfigured; if it is not, we'll just bail out without doing
**  anything as if the command were undefined.
*/
int mung_files(NODE *n)
{
    FILER *tmp;
    char *mung_args;
    char *cmdlin;
    char cmdlog[256];
    char filename[SPEC_SIZE+1];
    char line[133],name[NAME_LONG_SIZE+1],timbuf[40];
    char *p;
    FILE *fp;
    DSTRUCTURE *d;
    int num_mung_args = 0;
    int screen;

/*
**	First, see if CSWING$USERCOMMAND is defined
*/
	if ((cmdlin = getenv("CSWING$USERCOMMAND")) == NULL)
	{
	    warning("No command defined in CSWING$USERCOMMAND");
	    return(FALSE);
	}
/*
**	Next, open the scratch file we need.
**	Use usual cswing$save area if possible.
*/
    fp = NULL;
    sprintf(filename,"SWING$SAVE:MUNG_XXXXXX.LIS.");
    p = mktemp(filename);
    if (p)
        fp = fopen(p,"w","fop=cbt","alq=8","mbc=8");
    if (fp == NULL)
    {
        sprintf(filename,"SYS$LOGIN:MUNG_XXXXXX.LIS.");
        p = mktemp(filename);
        if (p)
            fp = fopen(p,"w","fop=cbt","alq=8","mbc=8");
        if (fp == NULL)
        {
            sprintf(junk,"Cannot create file %s",filename);
            warning(junk);
            return(TRUE);
        }
    }
/* now the scratch file is open... */
    if (!files_selected)
    {
        /*
        **  mung the current file
        */
	sprintf(junk,"Operate on the current file Y/N [N]");
	if (!confirm_user(junk,"Y")) return(FALSE);
        if (access(curr_file_ptr->filename,0) == 0)
        {
            tmp = current_file_ptr(curr_file_no);
	    fprintf(fp,"%s\n",tmp->filename);
	    num_mung_args++;
        }
        else
        {
            sprintf(junk,"%s is not accessable",curr_file_ptr->filename);
            warning(junk);
            return(FALSE);
        }
    }
    else
    {
        /*
        **  confirm intention to mung the selected files
        */
        if (files_selected == 1)
            sprintf(junk,"Operate on the selected file Y/N [N]");
        else
            sprintf(junk,"Operate on the %d selected files Y/N [N]",
                    files_selected);
        if (!confirm_user(junk,"Y"))
	{
	    sprintf(junk,"Operate on the current file Y/N [N]");
	    if (!confirm_user(junk,"Y")) return(FALSE);
	    /*
	    **  mung the current file
	    */
	    if (access(curr_file_ptr->filename,0) == 0)
	    {
		tmp = current_file_ptr(curr_file_no);
		fprintf(fp,"%s\n",tmp->filename);
		num_mung_args++;
	    }
	    else
	    {
		sprintf(junk,"%s is not accessable",curr_file_ptr->filename);
		warning(junk);
		return(FALSE);
	    }
	}
	else
	{
        /*
        **  mung the selected files
        */
        tmp = first_file_ptr;
        while(tmp)
        {
            if (tmp->rendition)
            {
                /*
                **  yup, this one's been chosen
                */
                if (access(tmp->filename,0) == 0)
                {
		fprintf(fp,"%s\n",tmp->filename);
		num_mung_args++;
                }
                else
                {
                sprintf(junk,"%s is not accessable",tmp->filename);
                warning(junk);
                }
            }
            tmp = tmp->forw_ptr;
        }
	}
    }

    /*
    **  if no files were accessible, return
    **  else do it
    */
/* first close the scratch file however... */
    fclose(fp);
    if (!num_mung_args){
        delete_a_file(p); /* delete the scratch file when done */
        return(FALSE);
    }
    else
    {
        smg$save_physical_screen(&paste,&screen);
        disable_broadcasts();

/* Fire off a command (trns of logical CSWING$USERCOMMAND)
   with our file spec as input */
	strcat(cmdlin," ");
	strcat(cmdlin,p);
        system(cmdlin);

        delete_a_file(p); /* delete the scratch file when done */

        enable_broadcasts();
        smg$restore_physical_screen(&paste,&screen);
    }

    return(TRUE);
}
